import sympy


class Assignment:
    """Representation of variable assignment
       similar to sympy.codegen.Assignment
    """
    def __init__(self, symbol, expression):
        """ symbol can be either string or sympy symbol
            symbol can be a string and a real symbol will be created
        """
        try:
            symbol.is_Symbol
            self.symbol = symbol
        except AttributeError:
            self.symbol = sympy.Symbol(symbol, real=True)
        self.expression = expression

    def subs(self, old, new):
        """Substitute old into new of rhs. Inplace
        """
        self.expression = self.expression.subs(old, new)

    @property
    def free_symbols(self):
        symbols = {self.symbol}
        symbols |= self.expression.free_symbols
        return symbols

    def __str__(self):
        return f'{self.symbol} := {self.expression}'


class Compartment:
    def __init__(self, name, input, output, amount=sympy.Function('A', real=True),
                 volume=sympy.Symbol('V'), idv=sympy.Symbol('t', real=True)):
        self.name = name
        self.input = input
        self.output = output
        self.amount = amount
        self.idv = idv
        self.volume = volume

    @property
    def free_symbols(self):
        return self.input.free_symbols | self.output.free_symbols | self.amount.free_symbols | \
               self.volume.free_symbols

    def de(self):
        """Differential equation"""
        dAdt = sympy.Derivative(self.amount(self.idv), self.idv)
        rhs = -self.output.rate * self.amount(self.idv)
        # Initial condition (second in tuble) depends on dose.
        # 0 means time relative to this dose. Should perhaps rather be "TIME"
        return sympy.Eq(dAdt, rhs), {self.amount(0): self.input.data_label}

    def __repr__(self):
        return f'{self.amount} := compartment({self.name}, input={self.input}, ' \
               f'output={self.output}, amount={self.amount}, volume={self.volume})'


class IVAbsorption:
    def __init__(self, data_label):
        self.data_label = data_label

    @property
    def free_symbols(self):
        return set()

    def __repr__(self):
        return f'IVAbsorption({self.data_label})'


class Elimination:
    def __init__(self, rate):
        self.rate = rate

    @property
    def free_symbols(self):
        return self.rate.free_symbols

    def __repr__(self):
        return f'Elimination({self.rate})'


class ModelStatements(list):
    """A list of sympy statements describing the model
    """
    @property
    def free_symbols(self):
        """Get a set of all free symbols"""
        symbols = set()
        for assignment in self:
            symbols |= assignment.free_symbols
        return symbols

    def subs(self, old, new):
        """Substitute old expression for new in all rhs of assignments"""
        for assignment in self:
            assignment.subs(old, new)

    def __str__(self):
        s = [str(assignment) for assignment in self]
        return '\n'.join(s)



"""Model specific tasks. Can be built into a model
"""


class Estimation:
    """A generic parameter estimation step
    """
    def __init__(self, evaluate=False, tool_options=None):
        self.evaluate = evaluate
        if tool_options is None:
            self.tool_options = []
        else:
            self.tool_options = tool_options


class FOCE:
    pass


class Uncertainty:
    """A generic parameter uncertainty step
    """
    pass



"""
=======================
Execution Run Directory
=======================

.. todo:: Create a derived class at :class:`pharmpy.api_nonmem.execute.run_directory`.

This module defines the :class:`.RunDirectory` class, which initialization creates a filesystem
directory that :class:`~pharmpy.tool.Tool` instances and :class:`~pharmpy.model.Model` methods can
utilize for input->output during evaluation, estimation & simulation tasks.

.. testcode::
    :pyversion: > 3.6

    from pharmpy.execute.run_directory import RunDirectory

Usage
-----

Creating a directory

    .. doctest::
        :pyversion: > 3.6

        >>> dir = RunDirectory('.')
        >>> print(list(dir.path.iterdir()))
        []

Defining clean options:

    .. doctest::
        :pyversion: > 3.6

        >>> dir.cleanlevel = 1
        >>> dir.clean_config(level=1, patterns=['*.txt', 'remove*'], rm_dirs=False)
        >>> files = ['remove_this', 'file.txt', 'keep_this', 'file.py']
        >>> for file in files:
        ...     open(dir.path / file, 'a').close()

Destroying object (triggering cleanup):

    .. doctest::
        :pyversion: > 3.6

        >>> path = dir.path
        >>> print(sorted([x.name for x in path.iterdir()]))
        ['file.py', 'file.txt', 'keep_this', 'remove_this']

    .. doctest::
        :pyversion: > 3.6

        >>> del dir
        >>> print([x.name for x in path.iterdir()])
        ['file.py', 'keep_this']

    .. doctest::
        :pyversion: > 3.6

        >>> (path / 'keep_this').unlink()
        >>> (path / 'file.py').unlink()
        >>> path.rmdir()

Definitions
-----------
"""

import itertools
import logging
import re
from pathlib import Path
from tempfile import TemporaryDirectory


class RunDirectory:
    """Execution run directory.
    Will contain input/output files for a job and handle cleanup. Is generated by
    :class:`~.engine.Engine` class (and tools).

    Args:
        parent: Parent (directory). If None, a temporary directory is created.
        name: (Base) name of directory (when parent is not None). (This) class name if False.
        template: Template for directory naming. '%s' substitutes for 'name' and '%d' for an int.

    Raises:
        FileExistsError: Constructed name is occupied (and '%d' is missing from 'template').

    Temporary directories are removed entirely on garbage collection. Default 'template' is
    '%s_dir%d'. First number in range (1, ∞) with a unique filesystem path is used.

    .. note::
        Expects caller to set :attr:`model` to iniate a data structure + file copy (if file
        representation is not already in directory), and *then execute*. Example::

            # run dir "takes" copy of model
            pheno = pysn.Model('pheno_real.mod')
            run_dir = RunDirectory('some/working/dir/', 'example_tool')
            run_dir.model = pheno

            # 'pheno' at original location & 'run_dir.model' copy in dir tree (safe to execute)
            run_dir.model != pheno
            run_dir.model.path != pheno.path
            run_dir.model.execute.estimate()
    """

    cleanlevel = 0
    """The active clean *level* of this directory."""

    def __init__(self, parent=None, name=None, template='%s_dir%d'):
        if not name:
            name = self.__class__.__name__

        if parent is None:
            self._tempdir = TemporaryDirectory(prefix='%s.' % (name,) if name else None)
            path = Path(self._tempdir.name)
        else:
            head = Path(str(parent)).resolve()
            if re.search(r'(?<!%)%d', template):
                for num in itertools.count(1):
                    path = head / (template % (name, num))
                    if not path.exists():
                        break
            else:
                path = head / (template % name)
            path.mkdir(exist_ok=False)

        self.path = path
        self.template = template
        self._model = None
        self._jobs = list()

    @property
    def name(self):
        """The name of this directory."""
        return self.path.name

    @property
    def model(self):
        """The current, main model of the directory (i.e. the target of execution)."""
        return self._model

    @model.setter
    def model(self, model):
        path = self.path / Path(model.path).name
        if path.is_file():
            raise ValueError('Model already exists, at: %r' % str(path))
        self._model = model.copy(dest=path)

    def bind_job(self, job):
        self._jobs += [job]

    def clean_config(self, level, patterns, rm_dirs=False):
        """Configure a cleanlevel.

        Args:
            level: The cleanup level.
            patterns: List of path-globbing patterns (removal targets).
            rm_dirs: Also delete non-empty directories matching 'patterns'?

        Level 0 should be reserved as the no-op.
        """
        num_prev = len(self.cleanlevels)
        if num_prev <= level:
            self._cleanlevels += [list() for _ in range(1 + num_prev - level)]
        self._cleanlevels[level] = {'glob': list(patterns), 'rm_dirs': bool(rm_dirs)}

    @property
    def cleanlevels(self):
        """The active cleanlevel *definitions* of this directory."""
        try:
            levels = self._cleanlevels
        except AttributeError:
            levels = self._cleanlevels = [[]]
        return levels

    def cleanup(self, level=None):
        """Delete files/dirs up to clean level (is run after completion).

        Args:
            level: The clean level to target. If None, (class) default clean level is used.
        """

        log = logging.getLogger(__name__)
        for job in self._jobs:
            if not job.done:
                log.warning('Job still running (%r): blocking cleanup of %r!', job, self)
                job.wait()

        if level is None:
            level = self.cleanlevel
        levels = self.cleanlevels[:(1+level)]
        for child in self.path.iterdir():
            clean = dict()
            for level in levels:
                if not level:
                    continue
                if any(child.match(glob) for glob in level['glob']):
                    clean = level
                    break
            if not clean:
                continue

            if child.is_file() or child.is_symlink():
                child.unlink()
            elif child.is_dir() and clean['rm_dirs']:
                for rchild in child.iterdir():
                    rchild.unlink()
                try:
                    rchild.rmdir()
                except OSError:
                    pass

    def __repr__(self):
        args = ['%r' % str(self.path.parent), '%r' % str(self.name)]
        if self.template != '%s_dir%d':
            args += ['%r' % self.template]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(args))

    def __str__(self):
        return str(self.path)

    def __del__(self):
        if self.path.exists():
            self.cleanup()



import pandas as pd
import sympy

from pharmpy.data_structures import OrderedSet


class ParameterSet(OrderedSet):
    """A set of parameters

    Class representing a group of parameters. Usually all parameters in a model.
    This class give a ways of displaying, summarizing and manipulating
    more than one parameter at a time.

    Specific parameters can be found using indexing on the parameter name

    .. code-block:: python

        params['THETA(1)']
    """
    def __getitem__(self, index):
        for e in self:
            if e.name == index:
                return e
        raise KeyError(f'Parameter "{index}" does not exist')

    def __contains__(self, item):
        for e in self:
            if e.name == item:
                return True
        return False

    def summary(self):
        """Give a dataframe with a summary of all Parameters

        :returns: A dataframe with one row per parameter.
                  The columns are value, lower, upper and fix
                  Row Index is the names
        """
        symbols = [param.name for param in self]
        values = [param.init for param in self]
        lower = [param.lower for param in self]
        upper = [param.upper for param in self]
        fix = [param.fix for param in self]
        return pd.DataFrame({'value': values, 'lower': lower, 'upper': upper, 'fix': fix},
                            index=symbols)

    @property
    def inits(self):
        """Initial estimates of parameters as dict
        """
        return {p.name: p.init for p in self}

    @inits.setter
    def inits(self, init_dict):
        for name, value in init_dict.items():
            self[name].init = value

    @property
    def fix(self):
        """Fixedness of parameters as dict
        """
        return {p.name: p.fix for p in self}

    @fix.setter
    def fix(self, fix_dict):
        for name, value in fix_dict.items():
            self[name].fix = value

    def remove_fixed(self):
        """Remove all fixed parameters
        """
        fixed = [p for p in self if p.fix]
        self -= fixed

    def __repr__(self):
        if len(self) == 0:
            return "ParameterSet()"
        return self.summary().to_string()

    def _repr_html_(self):
        if len(self) == 0:
            return "ParameterSet()"
        else:
            return self.summary().to_html()


class Parameter:
    """A single parameter

        .. code-block::

            param = Parameter("TVCL", 0.005, lower=0)
            param.fix = True

       .. attribute:: name

           Name of the parameter

       .. attribute:: fix

           A boolean to indicate whether the parameter is fixed or not. Note that fixing a parameter
           will keep its bounds even if a fixed parameter is actually constriand to one single
           value. This is so that unfixing will take back the previous bounds.
    """
    def __init__(self, name, init, lower=None, upper=None, fix=False):
        self._init = init
        self.name = name
        self.fix = bool(fix)
        self._lower = -sympy.oo
        self._upper = sympy.oo
        if lower is not None:
            self.lower = lower
        if upper is not None:
            self.upper = upper

    @property
    def lower(self):
        """Lower bound of the parameter
        """
        return self._lower

    @lower.setter
    def lower(self, new_lower):
        if new_lower > self.init:
            raise ValueError(f'Lower bound {new_lower} cannot be greater than init {self.init}')
        self._lower = new_lower

    @property
    def upper(self):
        """Upper bound of the parameter
        """
        return self._upper

    @upper.setter
    def upper(self, new_upper):
        if new_upper < self.init:
            raise ValueError(f'Upper bound {new_upper} cannot be less than init {self.init}')
        self._upper = new_upper

    @property
    def init(self):
        """Initial parameter estimate
        """
        return self._init

    @init.setter
    def init(self, new_init):
        if new_init < self.lower or new_init > self.upper:
            raise ValueError(f'Initial estimate must be within the constraints of the parameter: '
                             f'{new_init} ∉ {sympy.pretty(sympy.Interval(self.lower, self.upper))}'
                             f'\nUnconstrain the parameter before setting an initial estimate.')
        self._init = new_init

    def unconstrain(self):
        """Remove all constraints of a parameter
        """
        self._lower = -sympy.oo
        self._upper = sympy.oo
        self.fix = False

    @property
    def parameter_space(self):
        """The parameter space set

           A fixed parameter will be constrained to one single value
           and non-fixed parameters will be constrained to an interval
           possibly open in one or both ends.
        """
        if self.fix:
            return sympy.FiniteSet(self.init)
        else:
            return sympy.Interval(self.lower, self.upper)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        """Two parameters are equal if they have the same name, init and constraints
        """
        return self.init == other.init and self.lower == other.lower and \
            self.upper == other.upper and self.name == other.name and self.fix == other.fix

    def __repr__(self):
        return f'Parameter("{self.name}", {self.init}, lower={self.lower}, upper={self.upper}, ' \
               f'fix={self.fix})'



import pytest
import sympy

from pharmpy.model import ModelSyntaxError


def S(x):
    return sympy.Symbol(x)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize('buf,results', [
    ('$OMEGA 1', [('OMEGA(1,1)', 1, 0, sympy.oo, False)]),
    ('$OMEGA   0.123 \n\n', [('OMEGA(1,1)', 0.123, 0, sympy.oo, False)]),
    ('$OMEGA   (0 FIX) ; CL', [('OMEGA(1,1)', 0, 0, sympy.oo, True)]),
    ('$OMEGA DIAG(2) 1 2 FIX', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,2)', 2, 0, sympy.oo, True),
        ]),
    ('$OMEGA 1 2 3', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,2)', 2, 0, sympy.oo, False),
        ('OMEGA(3,3)', 3, 0, sympy.oo, False),
        ]),
    ('$OMEGA 0.15 ;CL', [('OMEGA(1,1)', 0.15, 0, sympy.oo, False)]),
    ('$OMEGA 1 \n2 ; S   \n  3 ', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,2)', 2, 0, sympy.oo, False),
        ('OMEGA(3,3)', 3, 0, sympy.oo, False),
        ]),
    ('$OMEGA 2 SD', [('OMEGA(1,1)', 4, 0, sympy.oo, False)]),
    ('$OMEGA ;CO\n (VAR 3)', [('OMEGA(1,1)', 3, 0, sympy.oo, False)]),
    ('$OMEGA (1)x2', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,2)', 1, 0, sympy.oo, False),
        ]),
    ('$OMEGA BLOCK(2) 1 0.5 2', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,1)', 0.5, -sympy.oo, sympy.oo, False),
        ('OMEGA(2,2)', 2, 0, sympy.oo, False),
        ]),
    ('$OMEGA BLOCK(2) SAME', []),
    ('$OMEGA BLOCK SAME(3)', []),
    ('$OMEGA BLOCK(2) 1 (0.1)x2', [
        ('OMEGA(1,1)', 1, 0, sympy.oo, False),
        ('OMEGA(2,1)', 0.1, -sympy.oo, sympy.oo, False),
        ('OMEGA(2,2)', 0.1, 0, sympy.oo, False),
        ]),
    ('$OMEGA BLOCK(2) CHOLESKY 0.8 -0.3 0.7', [
        ('OMEGA(1,1)', 0.64, 0, sympy.oo, False),
        ('OMEGA(2,1)', -0.24, -sympy.oo, sympy.oo, False),
        ('OMEGA(2,2)', 0.58, 0, sympy.oo, False),
        ]),
    ('$OMEGA BLOCK(2) SD 0.8 -0.394 0.762 CORR', [
        ('OMEGA(1,1)', 0.64, 0, sympy.oo, False),
        ('OMEGA(2,1)', -0.2401824, -sympy.oo, sympy.oo, False),
        ('OMEGA(2,2)', 0.580644, 0, sympy.oo, False),
        ]),
    ('$OMEGA BLOCK(1)   1.5', [
        ('OMEGA(1,1)', 1.5, 0, sympy.oo, False),
        ]),
    ('$OMEGA  0.0258583  ;      V2\n'
     ';$OMEGA BLOCK(1) 0.0075 FIX    ;.02 ; IOC\n'
     ';$OMEGA BLOCK(1) SAME\n', [
        ('OMEGA(1,1)', 0.0258583, 0, sympy.oo, False),
        ]),
])
def test_parameters(parser, buf, results):
    recs = parser.parse(buf)
    rec = recs.records[0]
    pset, _, _ = rec.parameters(1, 1)
    assert len(pset) == len(results)
    for res in results:
        name = res[0]
        init = res[1]
        lower = res[2]
        upper = res[3]
        fix = res[4]
        param = pset[name]
        assert param.name == name
        assert pytest.approx(param.init, 0.00000000000001) == init
        assert param.lower == lower
        assert param.upper == upper
        assert param.fix == fix


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize('buf', [
    ('$OMEGA 0 '),
    ('$OMEGA DIAG(1) 1 SD VARIANCE'),
    ('$OMEGA SD BLOCK(2) 0.1 0.001 0.1 STANDARD'),
    ('$OMEGA CHOLESKY BLOCK(2) 0.1 VAR 0.001 \n  0.1 '),
])
def test_errors(parser, buf):
    recs = parser.parse(buf)
    rec = recs.records[0]
    with pytest.raises(ModelSyntaxError):
        pset, _, _ = rec.parameters(1, None)


def test_parameters_offseted(parser):
    rec = parser.parse("$OMEGA 1").records[0]
    pset, _, _ = rec.parameters(3, None)
    assert pset['OMEGA(3,3)'].init == 1

    rec = parser.parse('$OMEGA   BLOCK   SAME').records[0]
    pset, _, _ = rec.parameters(3, 2)
    assert len(pset) == 0


def test_update(parser):
    rec = parser.parse('$OMEGA 1').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 2
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2.0'

    rec = parser.parse('$OMEGA 1 SD').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2.0 SD'

    rec = parser.parse('$OMEGA (1)x3\n;FTOL').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,2)'].init = 4
    pset['OMEGA(3,3)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA (4.0)x3\n;FTOL'

    rec = parser.parse('$OMEGA (1)x2 2').records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1
    pset['OMEGA(2,2)'].init = 2
    pset['OMEGA(3,3)'].init = 0.5
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA (1) (2.0) 0.5'

    rec = parser.parse("$OMEGA DIAG(2) (1 VAR) (SD 2)").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1.5
    pset['OMEGA(2,2)'].init = 16
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA DIAG(2) (1.5 VAR) (SD 4.0)'

    rec = parser.parse("$OMEGA BLOCK(2) 1 2 4").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 7
    pset['OMEGA(2,1)'].init = 0.5
    pset['OMEGA(2,2)'].init = 8
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 7.0 0.5 8.0'

    rec = parser.parse("$OMEGA BLOCK(2)\n SD 1 0.5 ;COM \n 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 0.25
    pset['OMEGA(2,2)'].init = 9
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2)\n SD 2.0 0.25 ;COM \n 3.0\n'

    rec = parser.parse("$OMEGA CORR BLOCK(2)  1 0.5 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 1
    pset['OMEGA(2,2)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA CORR BLOCK(2)  4.0 0.25 4.0\n'

    rec = parser.parse("$OMEGA CORR BLOCK(2)  1 0.5 SD 1\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 4
    pset['OMEGA(2,1)'].init = 1
    pset['OMEGA(2,2)'].init = 4
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA CORR BLOCK(2)  2.0 0.25 SD 2.0\n'

    rec = parser.parse("$OMEGA BLOCK(2) 1 0.1 1\n CHOLESKY").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 0.64
    pset['OMEGA(2,1)'].init = -0.24
    pset['OMEGA(2,2)'].init = 0.58
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 0.8 -0.3 0.7\n CHOLESKY'

    rec = parser.parse("$OMEGA BLOCK(2) (1)x3\n").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 0.64
    pset['OMEGA(2,1)'].init = -0.24
    pset['OMEGA(2,2)'].init = 0.58
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 0.64 -0.24 0.58\n'

    rec = parser.parse("$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].init = 1
    pset['OMEGA(2,1)'].init = 0.2
    pset['OMEGA(2,2)'].init = 3
    pset['OMEGA(3,1)'].init = 0.4
    pset['OMEGA(3,2)'].init = 0.5
    pset['OMEGA(3,3)'].init = 6
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(3) 1 ;CL\n0.2 3.0; V\n0.4 0.5 6.0; KA'

    rec = parser.parse("$OMEGA BLOCK(2) SAME").records[0]
    pset, _, _ = rec.parameters(3, 2)
    assert len(pset) == 0
    next_eta, prev_size = rec.update(pset, 3, 2)
    assert next_eta == 5
    assert prev_size == 2
    assert str(rec) == "$OMEGA BLOCK(2) SAME"

    rec = parser.parse("$OMEGA BLOCK SAME").records[0]
    pset, _, _ = rec.parameters(5, 4)
    assert len(pset) == 0
    next_eta, prev_size = rec.update(pset, 5, 4)
    assert next_eta == 9
    assert prev_size == 4
    assert str(rec) == "$OMEGA BLOCK SAME"


def test_update_fix(parser):
    rec = parser.parse("$OMEGA 2 FIX 4 (FIX 6)").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = False
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 4 (FIX 6)'

    rec = parser.parse("$OMEGA 2 4 6 ;STRAML").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = True
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 FIX 4 6 ;STRAML'
    pset['OMEGA(1,1)'].fix = False
    pset['OMEGA(2,2)'].fix = True
    pset['OMEGA(3,3)'].fix = True
    pset['OMEGA(3,3)'].init = 23
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA 2 4 FIX 23.0 FIX ;STRAML'

    rec = parser.parse("$OMEGA BLOCK(2) 1 .1 2 ;CLERMT").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = True
    # Cannot fix parts of block
    with pytest.raises(ValueError):
        rec.update(pset, 1, None)
    pset['OMEGA(2,1)'].fix = True
    pset['OMEGA(2,2)'].fix = True
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) FIX 1 .1 2 ;CLERMT'

    rec = parser.parse("$OMEGA BLOCK(2) 1 .1 FIX 1").records[0]
    pset, _, _ = rec.parameters(1, None)
    pset['OMEGA(1,1)'].fix = False
    pset['OMEGA(2,1)'].fix = False
    pset['OMEGA(2,2)'].fix = False
    rec.update(pset, 1, None)
    assert str(rec) == '$OMEGA BLOCK(2) 1 .1 1'


def test_random_variables(parser):
    rec = parser.parse("$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA").records[0]
    rvs, nxt, cov, zero_fix = rec.random_variables(1)
    assert nxt == 4
    assert len(rvs) == 3
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert rvs[2].name == 'ETA(3)'
    assert len(cov) == 9
    assert len(zero_fix) == 0

    rec = parser.parse("$OMEGA BLOCK(1) 1.5").records[0]
    rvs, nxt, cov, zero_fix = rec.random_variables(2)
    assert nxt == 3
    assert len(rvs) == 1
    assert rvs[0].name == 'ETA(2)'
    assert isinstance(rvs[0].pspace.distribution, sympy.stats.crv_types.NormalDistribution)
    assert cov == S('OMEGA(2,2)')
    assert len(zero_fix) == 0

    p = parser.parse("$OMEGA BLOCK(2) 1 0.01 1\n$OMEGA BLOCK(2) SAME\n")
    rec0 = p.records[0]
    rec1 = p.records[1]
    rvs, nxt, cov, zero_fix = rec0.random_variables(1)
    assert nxt == 3
    assert len(rvs) == 2
    assert rvs[0].name == 'ETA(1)'
    assert rvs[1].name == 'ETA(2)'
    assert len(cov) == 4
    assert len(zero_fix) == 0
    A = sympy.Matrix([[S('OMEGA(1,1)'), S('OMEGA(2,1)')], [S('OMEGA(2,1)'), S('OMEGA(2,2)')]])
    assert rvs[0].pspace.distribution.sigma == A
    rvs, nxt, cov, zero_fix = rec1.random_variables(nxt, cov)
    assert nxt == 5
    assert len(rvs) == 2
    assert rvs[0].name == 'ETA(3)'
    assert rvs[1].name == 'ETA(4)'
    assert len(cov) == 4
    assert len(zero_fix) == 0
    assert rvs[0].pspace.distribution.sigma == A

    rec = parser.parse("$OMEGA 0 FIX").records[0]
    rvs, _, _, zero_fix = rec.random_variables(1)
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)']

    rec = parser.parse("$OMEGA  BLOCK(2) FIX 0 0 0").records[0]
    rvs, _, _, zero_fix = rec.random_variables(1)
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)', 'ETA(2)']

    p = parser.parse("$OMEGA BLOCK(1) 0 FIX\n$OMEGA BLOCK(1) SAME")
    rec0 = p.records[0]
    rec1 = p.records[1]
    rvs, nxt, _, zero_fix = rec0.random_variables(1)
    assert nxt == 2
    assert len(rvs) == 0
    assert zero_fix == ['ETA(1)']
    rvs, nxt, _, zero_fix = rec1.random_variables(2, previous_cov='ZERO')
    assert nxt == 3
    assert len(rvs) == 0
    assert zero_fix == ['ETA(2)']

    rec = parser.parse("$OMEGA BLOCK SAME").records[0]
    A = sympy.Matrix([[1, 0.01], [0.01, 1]])
    rvs, _, _, _ = rec.random_variables(3, A)
    assert len(rvs) == 2
    assert list(rvs)[0].name == 'ETA(3)'
    assert list(rvs)[1].name == 'ETA(4)'



import enum
import itertools

import numpy as np
import symengine
import sympy
import sympy.stats as stats
from sympy.matrices import MatrixBase, MatrixExpr
from sympy.stats.rv import RandomSymbol

import pharmpy.math

from .data_structures import OrderedSet


class VariabilityLevel(enum.Enum):
    """ Representation of a variability level

       currently supports only IIV and RUV.
       Can be changed into something not-enum
       in future keeping the IIV and RUV as
       class singletons
    """
    IIV = enum.auto()
    RUV = enum.auto()


class MultivariateNormalDistribution(stats.joint_rv_types.MultivariateNormalDistribution):
    """A joint normal distribution that does not check for posdef.
       Needed because of sympy issue #18625
    """
    @staticmethod
    def check(mu, sigma):
        if mu.shape[0] != sigma.shape[0]:
            raise ValueError("Size of the mean vector and covariance matrix are incorrect.")


def Normal(name, mean, std):
    """sympy function patched to use our MultivariateNormalDistribution
    """
    if isinstance(mean, (list, MatrixBase, MatrixExpr)) and isinstance(std, (list, MatrixBase,
                                                                             MatrixExpr)):
        return stats.joint_rv_types.multivariate_rv(MultivariateNormalDistribution,
                                                    name, mean, std)
    return stats.crv_types.rv(name, stats.crv_types.NormalDistribution, (mean, std))


class JointDistributionSeparate(RandomSymbol):
    """One random variable in a joint distribution

       sympy can currently only represent the random variables in a joint distribution
       as one single indexed variable that cannot be separately named. This class
       makes separation possible.

       This class can probably not solve all issues with joint rvs, but it can at least
       handle seprate symbols, pass as a random variable for random_variables and lead back
       to its pspace.
    """
    def __new__(cls, name, joint_symbol):
        return super().__new__(cls, sympy.Symbol(name), joint_symbol.pspace)


def JointNormalSeparate(names, mean, cov):
    """Conveniently create a joint normal distribution and create separate random variables
    """
    x = Normal('__DUMMY__', mean, cov)
    rvs = [JointDistributionSeparate(name, x) for name in names]
    return rvs


class RandomVariables(OrderedSet):
    """An ordered set of random variables

       currently separately named jointrvs are not supported in sympy
       (i.e. it is not possible to do [eta1, eta2] = Normal(...))
       Use JointDistributionSeparate as a workaround
       Joints must come in the correct order
    """
    nletter = '𝒩 '

    @staticmethod
    def _left_parens(height):
        """Return an array containing each row of a large parenthesis
           used for pretty printing
        """
        a = ['⎧']
        for _ in range(height - 2):
            a.append('⎪')
        a.append('⎩')
        return a

    @staticmethod
    def _right_parens(height):
        """Return an array containing each row of a large parenthesis
           used for pretty printing
        """
        a = ['⎫']
        for _ in range(height - 2):
            a.append('⎪')
        a.append('⎭')
        return a

    @staticmethod
    def _normal_definition_string(rv):
        """Provide a array of pretty strings for the definition of a Normal random variable
           This should ideally be available from sympy.
        """
        dist = rv.pspace.distribution
        return [f'{sympy.pretty(rv, wrap_line=False)}'
                f'~ {RandomVariables.nletter}({sympy.pretty(dist.mean, wrap_line=False)}, '
                f'{sympy.pretty(dist.std**2, wrap_line=False)})']

    @staticmethod
    def _joint_normal_definition_string(rvs):
        """Provide an array of pretty strings for the definition of a Joint Normal random variable
        """
        dist = rvs[0].pspace.distribution
        name_vector = sympy.Matrix(rvs)
        name_strings = sympy.pretty(name_vector, wrap_line=False).split('\n')
        mu_strings = sympy.pretty(dist.mu, wrap_line=False).split('\n')
        sigma_strings = sympy.pretty(dist.sigma, wrap_line=False).split('\n')
        mu_height = len(mu_strings)
        sigma_height = len(sigma_strings)
        max_height = max(mu_height, sigma_height)

        left_parens = RandomVariables._left_parens(len(name_strings))
        right_parens = RandomVariables._right_parens(len(name_strings))

        # Pad the smaller of the matrices
        if mu_height != sigma_height:
            to_pad = mu_strings if mu_strings < sigma_strings else sigma_strings
            num_lines = abs(mu_height - sigma_height)
            padding = ' ' * len(to_pad[0])
            for i in range(0, num_lines):
                if i // 2 == 0:
                    to_pad.append(padding)
                else:
                    to_pad.insert(0, padding)

        central_index = max_height // 2
        res = []
        enumerator = enumerate(zip(name_strings, left_parens, mu_strings, sigma_strings,
                                   right_parens))
        for i, (name_line, lpar, mu_line, sigma_line, rpar) in enumerator:
            if i == central_index:
                res.append(name_line + f' ~ {RandomVariables.nletter}' + lpar +
                           mu_line + ', ' + sigma_line + rpar)
            else:
                res.append(name_line + '     ' + lpar + mu_line +
                           '  ' + sigma_line + rpar)

        return res

    def __getitem__(self, index):
        if isinstance(index, int):
            for i, e in enumerate(self):
                if i == index:
                    return e
        else:
            for e in self:
                if e == index or e.name == index:
                    return e
        raise KeyError(f'Random variable "{index}" does not exist')

    def __repr__(self):
        """ Give a nicely formatted view of the definitions of all
            random variables.
        """
        res = ''
        for rvs, dist in self.distributions():
            if isinstance(dist, stats.crv_types.NormalDistribution):
                lines = RandomVariables._normal_definition_string(rvs[0])
            elif isinstance(dist, stats.joint_rv_types.MultivariateNormalDistribution):
                lines = RandomVariables._joint_normal_definition_string(rvs)
            res += '\n'.join(lines) + '\n'
        return res

    def distributions(self, level=None):
        """Iterate with one entry per distribution instead of per random variable.

           level - only iterate over random variables of this variability level
        """
        i = 0
        while i < len(self):
            rv = self[i]
            dist = rv.pspace.distribution
            if isinstance(dist, stats.crv_types.NormalDistribution):
                i += 1
                if level is None or level == rv.variability_level:
                    yield [rv], dist
            else:       # Joint Normal
                n = self[i].pspace.distribution.sigma.rows
                rvs = [self[k] for k in range(i, i + n)]
                i += n
                if level is None or level == rv.variability_level:
                    yield rvs, dist

    def iiv_variance_parameters(self):
        parameters = []
        for rvs, dist in self.distributions(level=VariabilityLevel.IIV):
            if len(rvs) == 1:
                parameters.append(dist.std ** 2)
            else:
                parameters += list(dist.sigma.diagonal())
        return parameters

    def merge_normal_distributions(self, fill=0):
        """Merge all normal distributed rvs together into one joint normal

           Set new covariances (and previous 0 covs) to 'fill'
        """
        non_altered = []
        means = []
        blocks = []
        names = []
        for rvs, dist in self.distributions():
            names.extend([rv.name for rv in rvs])
            if isinstance(dist, stats.crv_types.NormalDistribution):
                means.append(dist.mean)
                blocks.append(sympy.Matrix([dist.std**2]))
            elif isinstance(dist, stats.joint_rv_types.MultivariateNormalDistribution):
                means.extend(dist.mu)
                blocks.append(dist.sigma)
            else:
                non_altered.extend(rvs)
        if names:
            if len(blocks) > 1:
                # Need special case for len(blocks) == 1 because of sympy 1.5.1 bug #18618
                M = sympy.BlockDiagMatrix(*blocks)
                M = sympy.Matrix(M)
            else:
                M = sympy.Matrix(blocks[0])
            if fill != 0:
                for row, col in itertools.product(range(M.rows), range(M.cols)):
                    if M[row, col] == 0:
                        M[row, col] = fill
            new_rvs = JointNormalSeparate(names, means, M)
            self.__init__(new_rvs + non_altered)

    def copy(self):
        # Special copy because separated joints need special treatment
        new_rvs = RandomVariables()
        for rvs, dist in self.distributions():
            if len(rvs) == 1:
                new_rvs.add(rvs[0].copy())
            else:
                new_rvs.update(JointNormalSeparate([rv.name for rv in rvs], dist.mu, dist.sigma))
        return rvs

    def validate_parameters(self, parameter_values, use_cache=False):
        """ Validate a dict or Series of parameter values

            Currently checks that all covariance matrices are posdef
            use_cache for using symengine cached matrices
        """
        if use_cache and not hasattr(self, '_cached_sigmas'):
            self._cached_sigmas = {}
            for rvs, dist in self.distributions():
                if len(rvs) > 1:
                    sigma = dist.sigma
                    a = [[symengine.Symbol(e.name) for e in sigma.row(i)]
                         for i in range(sigma.rows)]
                    A = symengine.Matrix(a)
                    self._cached_sigmas[rvs[0]] = A

        for rvs, dist in self.distributions():
            if len(rvs) > 1:
                if not use_cache:
                    sigma = dist.sigma.subs(dict(parameter_values))
                    # Switch to numpy here. Sympy posdef check is problematic
                    # see https://github.com/sympy/sympy/issues/18955
                    if not sigma.free_symbols:
                        a = np.array(sigma).astype(np.float64)
                        if not pharmpy.math.is_posdef(a):
                            return False
                else:
                    sigma = self._cached_sigmas[rvs[0]]
                    replacement = {}
                    # Following because https://github.com/symengine/symengine/issues/1660
                    for param in dict(parameter_values):
                        replacement[symengine.Symbol(param)] = parameter_values[param]
                    sigma = sigma.subs(replacement)
                    if not sigma.free_symbols:      # Cannot validate since missing params
                        a = np.array(sigma).astype(np.float64)
                        if not pharmpy.math.is_posdef(a):
                            return False
        return True

    def nearest_valid_parameters(self, parameter_values):
        """ Force parameter values into being valid

            As small changes as possible

            returns an updated parameter_values
        """
        nearest = parameter_values.copy()
        for rvs, dist in self.distributions():
            if len(rvs) > 1:
                sigma = dist.sigma.subs(dict(parameter_values))
                A = np.array(sigma).astype(np.float64)
                B = pharmpy.math.nearest_posdef(A)
                if B is not A:
                    for row in range(len(A)):
                        for col in range(row + 1):
                            nearest[dist.sigma[row, col].name] = B[row, col]
        return nearest


# pharmpy sets a parametrization attribute to the sympy distributions
# It will currently not affect the sympy distribution itself just convey the information
# The distribution class itself need to change to accomodate pdf etc.
# Could have different distributions for different parametrization, but would like to be able
# to convert between them. Ideally also support arbitrary parametrizations.
# Classes will be used by ModelfitResults to be able to reparametrize results without
# reparametrizing the whole model.

# For now simply use these for the results object to set proper parametrization of parameter.


class NormalParametrizationVariance:
    name = 'variance'
    distribution = stats.crv_types.NormalDistribution

    def __init__(self, mean, variance):
        pass


class NormalParametrizationSd:
    name = 'sd'
    distribution = stats.crv_types.NormalDistribution

    def __init__(self, mean, sd):
        pass


class MultivariateNormalParametrizationCovariance:
    name = 'covariance'
    distribution = stats.joint_rv_types.MultivariateNormalDistribution

    def __init__(self, mean, variance):
        pass


class MultivariateNormalParametrizationSdCorr:
    name = 'sdcorr'
    distribution = stats.joint_rv_types.MultivariateNormalDistribution

    def __init__(self, mean, sd, corr):
        pass



from __future__ import unicode_literals

import os


extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.githubpages',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'jupyter_sphinx',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'Pharmpy'
year = '2018-2020'
authors = ['the Pharmpy development team']
copyright = '{0}; {1}'.format(year, ', '.join(authors))
version = release = '0.3.0'

pygments_style = 'trac'
templates_path = ['.']

html_theme = "pydata_sphinx_theme"

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

# if this is True, todo and todolist produce output, else they produce nothing
todo_include_todos = True

# inheritance_graph_attrs = dict()
# inheritance_node_attrs = dict(font='Palatino', color='gray50', fontcolor='black')
# inheritance_edge_attrs = dict(color='maroon')
# graphviz_output_format = 'svg'



import itertools
from pathlib import Path

import altair as alt
import numpy as np
import pandas as pd
import symengine
import sympy

from pharmpy import Model
from pharmpy.data import ColumnType
from pharmpy.math import conditional_joint_normal, is_posdef
from pharmpy.parameter_sampling import sample_from_covariance_matrix, sample_individual_estimates
from pharmpy.random_variables import VariabilityLevel
from pharmpy.results import Results


class FREMResults(Results):
    """FREM Results class

        What follows is a description on how the FREM results are stored.
        See :py:mod:`pharmpy.methods.frem` for a description of how the results are calculated.

        .. attribute:: covariate_baselines

            DataFrame with the baseline covariate values used in the analysis for each indiviual.
            Index is ID. One column for each covariate.

            .. code-block::

                      WGT  APGR
                ID
                1.0   1.4   7.0
                2.0   1.5   9.0
                3.0   1.5   6.0
                4.0   0.9   6.0
                5.0   1.4   7.0
                6.0   1.2   5.0
                7.0   1.0   5.0
                8.0   1.2   7.0
                9.0   1.4   8.0
                ...   ...   ...


        .. attribute:: covariate_effects

            DataFrame with covariate effects. Index is parameter, covariate and condition where
            condition can be either 5th or 95th. Columns are p5, mean and p95 for the 5th
            percentile of the effect, the estimated effect and the 95th percentile of the effect
            respectively. Effect sizes are in fractions. Example:

            .. code-block::

                                                     p5      mean       p95
                parameter covariate condition
                ETA(1)    WGT       5th        0.901869  0.972920  1.051093
                                    95th       0.903849  1.064605  1.233111
                          APGR      5th        1.084238  1.248766  1.413923
                                    95th       0.848297  0.901919  0.962312
                ETA(2)    WGT       5th        0.942450  1.004400  1.068390
                                    95th       0.874409  0.995763  1.127777
                          APGR      5th        0.832008  0.924457  1.021307
                                    95th       0.990035  1.039444  1.091288

        .. attribute:: covariate_effects_plot

            Plot of the covariate effects generated by :py:meth:`plot_covariate_effects`

        .. attribute:: covariate_statistics

            DataFrame with summary statistics of the covariate baselines. Index is covariates.
            Columns are p5 with 95th percentile, mean, p95 with 95th percentile, stdev, ref with
            the used reference value, categorical and other with the non-reference value for
            categorical covariates. Example:

            .. code-block

                                        p5      mean  p95     stdev       ref  categorical  other
                            covariate
                            WGT        0.7  1.525424  3.2  0.704565  1.525424        False    NaN
                            APGR       1.0  6.423729  9.0  2.237636  6.423729        False    NaN

        .. attribute:: individual_effects

            DataFrame with individual covariate effects. Index is ID and parameter. Columns
            are observed, p5 and p95 for the observed individual effect, 5th and 95th
            percentiles of the estimated individual effects respectively. Example:

            .. code-block::

                                observed        p5       p95
                ID   parameter
                1.0  ETA(1)     0.973309  0.946282  0.982391
                     ETA(2)     1.009411  0.995276  1.024539
                2.0  ETA(1)     0.911492  0.832168  0.941760
                     ETA(2)     1.036200  0.990081  1.091438
                3.0  ETA(1)     1.013772  1.007555  1.028463
                ...                  ...       ...       ...
                57.0 ETA(2)     0.987500  0.942709  1.031111
                58.0 ETA(1)     0.939409  0.883782  0.956792
                     ETA(2)     1.023321  0.993543  1.057408
                59.0 ETA(1)     0.992578  0.952785  1.027261
                     ETA(2)     0.999220  0.968931  1.033819

        .. attribute:: individual_effects_plot

            Plot of the individual effects generated by :py:meth:`plot_individual_effects`

        .. attribute:: unexplained_variability

            DataFrame with remaining unexplained variability. Index is parameter and covariate.
            Covariate is none for no covariates and all for all covariates. Example:

            .. code-block::

                                     sd_observed    sd_5th   sd_95th
                parameter covariate
                ETA(1)    none          0.195327  0.221382  0.298465
                          WGT           0.194527  0.218385  0.292261
                          APGR          0.182497  0.202192  0.279766
                          all           0.178956  0.197282  0.268090
                ETA(2)    none          0.158316  0.152510  0.210044
                          WGT           0.158313  0.148041  0.207276
                          APGR          0.155757  0.149037  0.203477
                          all           0.155551  0.144767  0.201658

        .. attribute:: unexplained_variability_plot

            Plot of the unexplained variability generated by :py:meth:`plot_unexplained_variability`

    """
    rst_path = Path(__file__).parent / 'report.rst'

    def __init__(self, covariate_effects=None, individual_effects=None,
                 unexplained_variability=None, covariate_statistics=None,
                 covariate_effects_plot=None, individual_effects_plot=None,
                 unexplained_variability_plot=None, covariate_baselines=None):
        # FIXME: Lots of boilerplate code ahead. Could be simplified with python 3.7 dataclass
        #        or namedtuple
        self.covariate_effects = covariate_effects
        self.individual_effects = individual_effects
        self.unexplained_variability = unexplained_variability
        self.covariate_statistics = covariate_statistics
        self.covariate_effects_plot = covariate_effects_plot
        self.individual_effects_plot = individual_effects_plot
        self.unexplained_variability_plot = unexplained_variability_plot
        self.covariate_baselines = covariate_baselines

    def to_dict(self):
        return {'covariate_effects': self.covariate_effects,
                'individual_effects': self.individual_effects,
                'unexplained_variability': self.unexplained_variability,
                'covariate_statistics': self.covariate_statistics,
                'covariate_effects_plot': self.covariate_effects_plot,
                'individual_effects_plot': self.individual_effects_plot,
                'unexplained_variability_plot': self.unexplained_variability_plot,
                'covariate_baselines': self.covariate_baselines}

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def add_plots(self):
        self.covariate_effects_plot = self.plot_covariate_effects()
        self.individual_effects_plot = self.plot_individual_effects()
        self.unexplained_variability_plot = self.plot_unexplained_variability()

    def plot_covariate_effects(self):
        """Plot covariate effects
        """
        ce = (self.covariate_effects - 1) * 100
        cov_stats = pd.melt(self.covariate_statistics.reset_index(), var_name='condition',
                            id_vars=['covariate'], value_vars=['p5', 'p95'])

        cov_stats = cov_stats.replace({'p5': '5th', 'p95': '95th'}).set_index(['covariate',
                                                                               'condition'])
        ce = ce.join(cov_stats, how='inner')
        # The left join reorders the index. Might be pandas bug.
        ce = ce.reorder_levels(['parameter', 'covariate', 'condition'])

        param_names = list(ce.index.get_level_values('parameter').unique())
        plots = []

        for parameter in param_names:
            df = ce.xs(parameter, level=0)
            df = df.reset_index()

            error_bars = alt.Chart(df).mark_errorbar(ticks=True).encode(
                x=alt.X('p5:Q', title='Effect size in percent', scale=alt.Scale(zero=False)),
                x2=alt.X2('p95:Q'),
                y=alt.Y('condition:N', title=None),
            )

            rule = alt.Chart(df).mark_rule(strokeDash=[10, 4], color='gray').encode(
                x=alt.X('xzero:Q')
            ).transform_calculate(xzero="0")

            points = alt.Chart(df).mark_point(filled=True, color='black').encode(
                x=alt.X('mean:Q'),
                y=alt.Y('condition:N'),
            )

            text = alt.Chart(df).mark_text(
                dy=-15,
                color="red"
            ).encode(
                x=alt.X("mean:Q"),
                y=alt.Y("condition:N"),
                text=alt.Text("value:Q")
            )

            plot = alt.layer(error_bars, rule, points, text, data=df, width=800, height=100).facet(
                columns=1.0,
                row=alt.Facet('covariate:N', title=None),
                title=f'{parameter}'
            )

            plots.append(plot)

        v = alt.vconcat(*plots).resolve_scale(x='shared')
        return v

    def plot_individual_effects(self):
        """Plot individual effects
        """
        covs = self.covariate_baselines
        ie = self.individual_effects.join(covs)
        param_names = list(ie.index.get_level_values('parameter').unique())
        ie = (ie - 1) * 100
        ie = ie.sort_values(by=['observed'])

        plots = []

        for parameter in param_names:
            df = ie.xs(parameter, level=1)

            id_order = list(df.index)
            id_order = [str(int(x)) for x in id_order]

            if len(df) > 20:
                id_order[10] = '...'

            df = df.reset_index()
            df['ID'] = df['ID'].astype(int).astype(str)

            error_bars = alt.Chart(df).mark_errorbar(ticks=True).encode(
                x=alt.X('p5:Q', title='Effect size in percent', scale=alt.Scale(zero=False)),
                x2=alt.X2('p95:Q'),
                y=alt.Y('ID:N', title='ID', sort=id_order),
                tooltip=['ID', 'p5', 'observed', 'p95'] + list(covs.columns),
            )

            rule = alt.Chart(df).mark_rule(strokeDash=[10, 2], color='gray').encode(
                x=alt.X('xzero:Q')
            ).transform_calculate(xzero="0")

            points = alt.Chart(df).mark_point(size=40, filled=True, color='black').encode(
                x=alt.X('observed:Q'),
                y=alt.Y('ID:N', sort=id_order),
            )

            plot = alt.layer(points, error_bars, rule, data=df, width=800,
                             title=f'Individuals for parameter {parameter}')
            if len(df) > 20:
                plot = plot.transform_window(
                        sort=[alt.SortField('observed', order='ascending')],
                        rank='row_number(observed)',
                    ).transform_window(
                         sort=[alt.SortField('observed', order='descending')],
                         nrank='row_number(observed)'
                    ).transform_filter(
                        'datum.rank <= 10 | datum.nrank <= 11'
                    ).transform_calculate(
                        ID="datum.nrank == 11 ? '...' : datum.ID",
                        p5="datum.nrank == 11 ? '...' : datum.p5",
                        p95="datum.nrank == 11 ? '...' : datum.p95",
                        observed="datum.nrank == 11 ? '...' : datum.observed",
                    )
            plots.append(plot)

        v = alt.vconcat(*plots).resolve_scale(x='shared')
        return v

    def plot_unexplained_variability(self):
        """Plot unexplained variability
        """
        uv = self.unexplained_variability
        param_names = list(uv.index.get_level_values('parameter').unique())
        cov_order = list(uv.index.get_level_values('covariate').unique())
        plots = []

        for parameter in param_names:
            df = uv.xs(parameter, level=0)
            df = df.reset_index()

            error_bars = alt.Chart(df).mark_errorbar(ticks=True).encode(
                x=alt.X('sd_5th:Q', title='SD of unexplained variability',
                        scale=alt.Scale(zero=False)),
                x2=alt.X2('sd_95th:Q'),
                y=alt.Y('covariate:N', title='covariate', sort=cov_order),
                tooltip=['sd_5th', 'sd_observed', 'sd_95th', 'covariate'],
            )

            rule = alt.Chart(df).mark_rule(strokeDash=[10, 2], color='gray').encode(
                x=alt.X('xzero:Q')
            ).transform_calculate(xzero="0")

            points = alt.Chart(df).mark_point(size=40, filled=True, color='black').encode(
                x=alt.X('sd_observed:Q'),
                y=alt.Y('covariate:N', sort=cov_order),
            )

            plot = alt.layer(points, error_bars, rule, data=df, width=800,
                             title=f'Unexplained variability on {parameter}')

            plots.append(plot)

        v = alt.vconcat(*plots).resolve_scale(x='shared')
        return v

    def __str__(self):
        start = 'Results from FREM'
        # method...
        # covs = f'Continuous covariates: {self.continuous}'
        # cats = f'Categorical covariates: {self.categorical}'
        # FIXME: General option object for this?
        # if self.rescale:
        #    resc = 'Rescaling was used'
        # else:
        #    resc = 'Rescaling was not used'
        ce = self.covariate_effects.to_string()
        ie = self.individual_effects.to_string()
        uv = self.unexplained_variability.to_string()
        cs = self.covariate_statistics.to_string()
        return f'{start}\n\nCovariate statistics\n{cs}\n\n' \
               f'Covariate effects\n{ce}\n\nIndividual effects\n{ie}\n\n' \
               f'Unexplained variability\n{uv}\n'


def calculate_results(frem_model, continuous, categorical, method=None, **kwargs):
    """Calculate FREM results

       :param method: Either 'cov_sampling' or 'bipp'
    """
    if method is None or method == 'cov_sampling':
        try:
            res = calculate_results_using_cov_sampling(frem_model, continuous,
                                                       categorical, **kwargs)
        except AttributeError:
            # Fallback to bipp
            res = calculate_results_using_bipp(frem_model, continuous, categorical)
    elif method == 'bipp':
        res = calculate_results_using_bipp(frem_model, continuous, categorical)
    else:
        raise ValueError(f'Unknown frem postprocessing method {method}')
    return res


def calculate_results_using_cov_sampling(frem_model, continuous, categorical, cov_model=None,
                                         force_posdef_samples=500, force_posdef_covmatrix=False,
                                         samples=1000, rescale=True):
    """Calculate the FREM results using covariance matrix for uncertainty

       :param cov_model: Take the parameter uncertainty covariance matrix from this model
                         instead of the frem model.
       :param force_posdef_samples: The number of sampling tries before stopping to use
                                    rejection sampling and instead starting to shift values so
                                    that the frem matrix becomes positive definite. Set to 0 to
                                    always force positive definiteness.
       :param force_posdef_covmatrix: Set to force the covariance matrix of the frem movdel or
                                      the cov model to be positive definite. Default is to raise
                                      in this case.
       :param samples: The number of parameter vector samples to use.
    """
    if cov_model is not None:
        uncertainty_results = cov_model.modelfit_results
    else:
        uncertainty_results = frem_model.modelfit_results

    _, dist = list(frem_model.random_variables.distributions(
        level=VariabilityLevel.IIV))[-1]
    sigma_symb = dist.sigma

    parameters = [s for s in frem_model.modelfit_results.parameter_estimates.index
                  if sympy.Symbol(s) in sigma_symb.free_symbols]
    parvecs = sample_from_covariance_matrix(frem_model,
                                            modelfit_results=uncertainty_results,
                                            force_posdef_samples=force_posdef_samples,
                                            force_posdef_covmatrix=force_posdef_covmatrix,
                                            parameters=parameters,
                                            n=samples)
    res = calculate_results_from_samples(frem_model, continuous, categorical, parvecs,
                                         rescale=rescale)
    return res


def calculate_results_from_samples(frem_model, continuous, categorical, parvecs, rescale=True):
    """Calculate the FREM results given samples of parameter estimates
    """
    n = len(parvecs)
    rvs, dist = list(frem_model.random_variables.distributions(
        level=VariabilityLevel.IIV))[-1]
    sigma_symb = dist.sigma
    parameters = [s for s in frem_model.modelfit_results.parameter_estimates.index
                  if sympy.Symbol(s) in sigma_symb.free_symbols]
    parvecs = parvecs.append(
            frem_model.modelfit_results.parameter_estimates.loc[parameters])

    df = frem_model.dataset
    covariates = continuous + categorical
    df.pharmpy.column_type[covariates] = ColumnType.COVARIATE
    covariate_baselines = df.pharmpy.covariate_baselines
    covariate_baselines = covariate_baselines[covariates]
    cov_means = covariate_baselines.mean()
    cov_modes = covariate_baselines.mode().iloc[0]      # Select first mode if more than one
    cov_others = pd.Series(index=cov_modes[categorical].index, dtype=np.float64)
    cov_stdevs = covariate_baselines.std()
    for _, row in covariate_baselines.iterrows():
        for cov in categorical:
            if row[cov] != cov_modes[cov]:
                cov_others[cov] = row[cov]
        if not cov_others.isna().values.any():
            break

    cov_refs = pd.concat((cov_means[continuous], cov_modes[categorical]))
    cov_5th = covariate_baselines.quantile(0.05, interpolation='lower')
    cov_95th = covariate_baselines.quantile(0.95, interpolation='higher')
    is_categorical = cov_refs.index.isin(categorical)

    res = FREMResults()

    res.covariate_statistics = pd.DataFrame({'p5': cov_5th, 'mean': cov_means,
                                             'p95': cov_95th, 'stdev': cov_stdevs,
                                             'ref': cov_refs, 'categorical': is_categorical,
                                             'other': cov_others}, index=covariates)
    res.covariate_statistics.index.name = 'covariate'

    ncovs = len(covariates)
    npars = sigma_symb.rows - ncovs
    nids = len(covariate_baselines)
    param_indices = list(range(npars))
    if rescale:
        scaling = np.diag(np.concatenate((np.ones(npars), cov_stdevs.values)))

    mu_bars_given_5th = np.empty((n, ncovs, npars))
    mu_bars_given_95th = np.empty((n, ncovs, npars))
    mu_id_bars = np.empty((n, nids, npars))
    original_id_bar = np.empty((nids, npars))
    variability = np.empty((n, ncovs + 2, npars))      # none, cov1, cov2, ..., all
    original_variability = np.empty((ncovs + 2, npars))

    # Switch to symengine for speed
    # Could also assume order of parameters, but not much gain
    sigma_symb = symengine.sympify(sigma_symb)
    parvecs.columns = [symengine.Symbol(colname) for colname in parvecs.columns]

    covbase = covariate_baselines.to_numpy()

    for sample_no, params in parvecs.iterrows():
        sigma = sigma_symb.subs(dict(params))
        sigma = np.array(sigma).astype(np.float64)
        if rescale:
            sigma = scaling @ sigma @ scaling
        if sample_no != 'estimates':
            variability[sample_no, 0, :] = np.diag(sigma)[:npars]
        else:
            original_variability[0, :] = np.diag(sigma)[:npars]
        for i, cov in enumerate(covariates):
            indices = param_indices + [i + npars]
            cov_sigma = sigma[indices][:, indices]
            cov_mu = np.array([0] * npars + [cov_refs[cov]])
            if cov in categorical:
                first_reference = cov_others[cov]
            else:
                first_reference = cov_5th[cov]
            mu_bar_given_5th_cov, sigma_bar = conditional_joint_normal(
                cov_mu, cov_sigma, np.array([first_reference]))
            if sample_no != 'estimates':
                mu_bar_given_95th_cov, _ = conditional_joint_normal(cov_mu, cov_sigma,
                                                                    np.array([cov_95th[cov]]))
                mu_bars_given_5th[sample_no, i, :] = mu_bar_given_5th_cov
                mu_bars_given_95th[sample_no, i, :] = mu_bar_given_95th_cov
                variability[sample_no, i + 1, :] = np.diag(sigma_bar)
            else:
                original_variability[i + 1, :] = np.diag(sigma_bar)

        for i in range(len(covariate_baselines)):
            row = covbase[i, :]
            id_mu = np.array([0] * npars + list(cov_refs))
            mu_id_bar, sigma_id_bar = conditional_joint_normal(id_mu, sigma, row)
            if sample_no != 'estimates':
                mu_id_bars[sample_no, i, :] = mu_id_bar
                variability[sample_no, -1, :] = np.diag(sigma_id_bar)
            else:
                original_id_bar[i, :] = mu_id_bar
                original_variability[ncovs + 1, :] = np.diag(sigma_id_bar)

    # Create covariate effects table
    mu_bars_given_5th = np.exp(mu_bars_given_5th)
    mu_bars_given_95th = np.exp(mu_bars_given_95th)

    means_5th = np.mean(mu_bars_given_5th, axis=0)
    means_95th = np.mean(mu_bars_given_95th, axis=0)
    q5_5th = np.quantile(mu_bars_given_5th, 0.05, axis=0)
    q5_95th = np.quantile(mu_bars_given_95th, 0.05, axis=0)
    q95_5th = np.quantile(mu_bars_given_5th, 0.95, axis=0)
    q95_95th = np.quantile(mu_bars_given_95th, 0.95, axis=0)

    df = pd.DataFrame(columns=['parameter', 'covariate', 'condition', 'p5', 'mean', 'p95'])
    param_names = [rv.name for rv in rvs][:npars]
    for param, cov in itertools.product(range(npars), range(ncovs)):
        if covariates[cov] in categorical:
            df = df.append({'parameter': param_names[param], 'covariate': covariates[cov],
                            'condition': 'other', 'p5': q5_5th[cov, param],
                            'mean': means_5th[cov, param], 'p95': q95_5th[cov, param]},
                           ignore_index=True)
        else:
            df = df.append({'parameter': param_names[param], 'covariate': covariates[cov],
                            'condition': '5th', 'p5': q5_5th[cov, param],
                            'mean': means_5th[cov, param], 'p95': q95_5th[cov, param]},
                           ignore_index=True)
            df = df.append({'parameter': param_names[param], 'covariate': covariates[cov],
                            'condition': '95th', 'p5': q5_95th[cov, param],
                            'mean': means_95th[cov, param], 'p95': q95_95th[cov, param]},
                           ignore_index=True)
    df.set_index(['parameter', 'covariate', 'condition'], inplace=True)
    res.covariate_effects = df

    # Create id table
    mu_id_bars = np.exp(mu_id_bars)
    original_id_bar = np.exp(original_id_bar)

    with np.testing.suppress_warnings() as sup:     # Would warn in case of missing covariates
        sup.filter(RuntimeWarning, "All-NaN slice encountered")
        id_5th = np.nanquantile(mu_id_bars, 0.05, axis=0)
        id_95th = np.nanquantile(mu_id_bars, 0.95, axis=0)

    df = pd.DataFrame(columns=['parameter', 'observed', 'p5', 'p95'])
    for curid, param in itertools.product(range(nids), range(npars)):
        df = df.append(pd.Series({'parameter': param_names[param],
                                  'observed': original_id_bar[curid, param],
                                  'p5': id_5th[curid, param],
                                  'p95': id_95th[curid, param]},
                                 name=covariate_baselines.index[curid]))
    df.index.name = 'ID'
    df = df.set_index('parameter', append=True)
    res.individual_effects = df

    # Create unexplained variability table
    sd_5th = np.sqrt(np.nanquantile(variability, 0.05, axis=0))
    sd_95th = np.sqrt(np.nanquantile(variability, 0.95, axis=0))
    original_sd = np.sqrt(original_variability)

    df = pd.DataFrame(columns=['parameter', 'covariate', 'sd_observed', 'sd_5th', 'sd_95th'])
    for par, cond in itertools.product(range(npars), range(ncovs + 2)):
        if cond == 0:
            condition = 'none'
        elif cond == ncovs + 1:
            condition = 'all'
        else:
            condition = covariates[cond - 1]
        df = df.append({'parameter': param_names[par], 'covariate': condition,
                        'sd_observed': original_sd[cond, par], 'sd_5th': sd_5th[cond, par],
                        'sd_95th': sd_95th[cond, par]}, ignore_index=True)
    df = df.set_index(['parameter', 'covariate'])
    res.unexplained_variability = df

    res.covariate_baselines = covariate_baselines

    return res


def calculate_results_using_bipp(frem_model, continuous, categorical, rescale=True, samples=2000):
    """Estimate a covariance matrix for the frem model using the BIPP method

       Bootstrap on the individual parameter posteriors
       Only the individual estimates, individual unvertainties and the parameter estimates
       are needed.

    """
    rvs, dist = list(frem_model.random_variables.distributions(
        level=VariabilityLevel.IIV))[-1]
    etas = [rv.name for rv in rvs]
    pool = sample_individual_estimates(frem_model, parameters=etas)
    ninds = len(pool.index.unique())
    ishr = frem_model.modelfit_results.individual_shrinkage
    lower_indices = np.tril_indices(len(etas))
    pop_params = np.array(dist.sigma).astype(str)[lower_indices]
    parameter_samples = np.empty((samples, len(pop_params)))
    remaining_samples = samples
    k = 0
    while k < remaining_samples:
        bootstrap = pool.sample(n=ninds, replace=True)
        ishk = ishr.loc[bootstrap.index]
        cf = (1 / (1 - ishk.mean())) ** (1/2)
        corrected_bootstrap = bootstrap * cf
        bootstrap_cov = corrected_bootstrap.cov()
        if not is_posdef(bootstrap_cov.to_numpy()):
            continue
        parameter_samples[k, :] = bootstrap_cov.values[lower_indices]
        k += 1
    frame = pd.DataFrame(parameter_samples, columns=pop_params)
    res = calculate_results_from_samples(frem_model, continuous, categorical, frame,
                                         rescale=rescale)
    return res


def psn_frem_results(path, force_posdef_covmatrix=False, force_posdef_samples=500, method=None):
    """ Create frem results from a PsN FREM run

        :param path: Path to PsN frem run directory
        :return: A :class:`FREMResults` object

    """
    path = Path(path)

    model_4_path = path / 'final_models' / 'model_4.mod'
    if not model_4_path.is_file():
        raise IOError(f'Could not find FREM model 4: {str(model_4_path)}')
    model_4 = Model(model_4_path)
    if model_4.modelfit_results is None:
        raise ValueError('Model 4 has no results')
    cov_model = None
    if method == 'cov_sampling':
        try:
            model_4.modelfit_results.covariance_matrix
        except Exception:
            model_4b_path = path / 'final_models' / 'model_4b.mod'
            try:
                model_4b = Model(model_4b_path)
            except FileNotFoundError:
                pass
            else:
                cov_model = model_4b

    with open(path / 'covariates_summary.csv') as covsum:
        covsum.readline()
        raw_cov_list = covsum.readline()

    all_covariates = raw_cov_list[1:].rstrip().split(',')
    nunique = model_4.dataset.pharmpy.baselines[all_covariates].nunique()
    continuous = list(nunique.index[nunique != 2])
    categorical = list(nunique.index[nunique == 2])

    # FIXME: Not introducing yaml parser in pharmpy just yet. Options should be collected
    # differently. Perhaps using json
    with open(path / 'meta.yaml') as meta:
        for row in meta:
            row = row.strip()
            if row.startswith('rescale: 1'):
                rescale = True
            elif row.startswith('rescale: 0'):
                rescale = False

    res = calculate_results(model_4, continuous, categorical, method=method,
                            force_posdef_covmatrix=force_posdef_covmatrix,
                            force_posdef_samples=force_posdef_samples, cov_model=cov_model,
                            rescale=rescale)
    return res



import os
import re
import shutil
import tempfile
from pathlib import Path
from urllib.request import urlopen

from bs4 import BeautifulSoup
from csscompressor import compress
from sphinx.application import Sphinx


class TemporaryDirectoryChanger:
    def __init__(self, path):
        self.path = path
        self.old_path = Path.cwd()

    def __enter__(self):
        os.chdir(self.path)
        return self

    def __exit__(self, *args):
        os.chdir(self.old_path)


def generate_report(rst_path, results_path):
    """Generate report from rst and results json
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmp_path = Path(tmpdirname)
        source_path = tmp_path / 'source'
        source_path.mkdir()
        shutil.copy(rst_path, source_path / 'results.rst')
        shutil.copy(results_path / 'results.json', source_path)

        conf_path = Path(__file__).parent

        # Change directory for results.json to be found
        with TemporaryDirectoryChanger(source_path):
            with open(os.devnull, 'w') as devnull:
                app = Sphinx(str(source_path), str(conf_path), str(tmp_path), str(tmp_path),
                             "singlehtml", status=devnull, warning=devnull)
                app.build()

        # Write missing altair css
        with open(tmp_path / '_static' / 'altair-plot.css', 'w') as dh:
            dh.write(""".vega-actions a {
    margin-right: 12px;
    color: #757575;
    font-weight: normal;
    font-size: 13px;
}

.vega-embed {
    margin-bottom: 20px;
    margin-top: 20px;
    width: 100%;
}
""")
        report_path = tmp_path / 'results.html'
        embed_css_and_js(tmp_path / 'results.html', report_path)
        shutil.copy(report_path, results_path)


def embed_css_and_js(html, target):
    """Embed all external css and javascript into an html
    """
    with open(html, 'r', encoding='utf-8') as sh:
        soup = BeautifulSoup(sh, features='lxml')

    scripts = soup.findAll("script", attrs={"src": True})

    for script in scripts:
        source = script.attrs['src']
        if source.startswith('http'):
            infile = urlopen(source)
            content = infile.read().decode('utf-8')
            infile.close()
        else:
            path = html.parent / source
            with open(path, 'r') as sh:
                content = sh.read()

        # Minification with jsmin didn't work
        tag = soup.new_tag('script')
        tag['type'] = 'text/javascript'
        tag.append(content)
        script.replace_with(tag)

    stylesheets = soup.findAll("link", attrs={"rel": "stylesheet"})

    for stylesheet in stylesheets:
        stylesheet_src = stylesheet.attrs['href']
        tag = soup.new_tag("style")
        tag['type'] = 'text/css'
        path = html.parent / stylesheet_src
        with open(path, 'r') as sh:
            content = sh.read()
        if '@import' in content:
            import_files = re.findall(r'@import\s+url\("([A-Za-z0-9.]+)"\)', content)
            for name in import_files:
                with open(path.parent / name, 'r') as import_file:
                    import_content = import_file.read()
                content = re.sub(r'@import\s+url\("' + name + r'"\);', import_content, content)
        minified_content = compress(content)
        tag.append(minified_content)
        stylesheet.replace_with(tag)

    with open(target, 'w', encoding='utf-8') as dh:
        dh.write(str(soup))




import csv
import random
import re
import string
from collections import namedtuple
from pathlib import Path

import pytest


@pytest.fixture(scope='session')
def testdata():
    """Test data (root) folder."""
    return Path(__file__).resolve().parent / 'testdata'


@pytest.fixture(scope='session')
def csv_read():
    """Load test data from CSV. Supports tuples."""

    _tuple_matcher = re.compile(r'^\((.*)\)$')

    def func(root, file, names=None):
        TestData = tuple
        if names:
            TestData = namedtuple('TestData', names)

        def descape(item):
            return item.replace('\\n', '\n')

        def tupleize(item):
            m = _tuple_matcher.match(item)
            if not m:
                return descape(item)
            if not m.group(1):
                return ()
            return tuple(descape(x.strip(' "')) for x in m.group(1).split(','))

        with open(Path(root, file), 'r') as f:
            dialect = csv.Sniffer().sniff(f.read(1024))
            f.seek(0)
            reader = csv.reader(f, dialect)
            return tuple(TestData(*tuple(map(tupleize, row))) for row in reader)

    return func


@pytest.fixture(scope='session')
def str_repr():
    """Debug print (string representations)."""

    def func(string):
        if not string:
            return '-- EMPTY --'
        return '"' + repr(string)[1:-1] + '"'
    return func


class _RandomData:
    """
    A collection of random data generators.

    Use to procedurally generate data for testing.

    Args:
        N (int): Number of values to cap generators on. Infinite if None.
        comment_char (str): Char to begin comments (and exclude from comment content).
    """

    comment_char = ';'
    comment_charset = (' '*25 + string.ascii_letters*5 + string.digits*3 + '\t'*3 +
                       string.punctuation).replace(comment_char, '')

    def __init__(self, N=None, comment_char=';'):
        self.history = []
        if comment_char != self.comment_char:
            self.comment_charset = self.comment_charset.replace(self.comment_char, comment_char)
            self.comment_char = comment_char
        self.N = N

    # -- methods producing random generators ---------------------------
    def pos_int(self, size=1E4):
        """Returns (generator for) non-negative integer x, where x <= size."""
        return self._gen(random.randint, 0, size)

    def int(self, size=1E4):
        """Returns (generator for) integer x, where -size < x < size."""
        return self._gen(random.randint, -size, size)

    def float(self, size=1E9):
        """Returns (generator for) floating point num x, where x ~ normal_dist(0, sd=size)."""
        return self._gen(random.normalvariate, 0, size)

    def str_num(self, N=0, sci_fmt=('%E', '%.0E')):
        """Returns (generator for) tuple (str, num) from mixed int/float (convenience function).

        Args:
            N (int): Generate 1 tuple (str, num) if 0. Else, sorted (nested) tuple of 'N' tuples.
            sci_fmt (tuple): Format to randomly apply when formatting floats.
        """

        def f(sci_fmt, scalar, n):
            def g():
                if random.getrandbits(1):
                    if random.getrandbits(1):
                        val = random.randint(0, 9)
                    else:
                        val = random.randint(-1E5, 1E5)
                    s = str(val)
                else:
                    fmt = random.choice(sci_fmt)
                    if random.getrandbits(1):
                        val = random.normalvariate(0, 1)
                    else:
                        val = random.normalvariate(0, 10**(random.randint(-50, 50)))
                    s = fmt % (val,)
                    val = float(s)
                return val, s

            if scalar:
                return g()
            else:
                sort = sorted((g() for _ in range(n)), key=lambda x: x[0])
                return tuple(sort)

        return self._gen(f, sci_fmt, not N, N)

    def bool(self):
        """Returns (generator for) bool."""
        return self._gen(lambda: bool(random.getrandbits(1)))

    def choice(self, *choices):
        """Returns (generator for) choice (in set from args)."""
        return self._gen(lambda x: random.choice(x), choices)

    def biased_choice(self, *bias_choice):
        """
        Returns (generator for) biased choice (in set from args).

        Each arg is a tuple (bias, choice) where bias is the relative frequency (float).
        """

        biases, choices = zip(*bias_choice)
        base = min(biases)*100
        rep = tuple(round((b*100)/base) for b in biases)
        choices = [[ch]*rep for ch, rep in zip(choices, rep)]
        choices = [x for lst in choices for x in lst]
        return self._gen(lambda x: random.choice(x), choices)

    def comment(self, maxlen=30):
        """
        Returns (generator for) comment (uses :attr:`comment_char` and :attr:`comment_charset`).

        Args:
            maxlen (int): Maximum length of comment.
        """

        def f(st, ch, _max):
            comment = st + random.choice(['', ' '])
            if not bool(random.getrandbits(2)):
                return comment
            return comment + ''.join(random.choice(ch) for _ in range(1, random.randrange(_max)))
        return self._gen(f, self.comment_char, self.comment_charset, maxlen)

    def pad(self, maxlen=5, nl=False):
        """
        Returns (generator for) whitespace padding/empty str 50/50.

        Args:
            maxlen (int): Maximum length of str.
            nl (bool): Allow line endings (half as likely as ' ').
        """

        def f(x):
            return random.choice(chars)*random.choice([0]*x + list(range(1, x)))
        if nl:
            chars = ' \n '
        else:
            chars = ' '
        return self._gen(f, maxlen)

    # -- private methods -----------------------------------------------
    def _gen(self, f, *args, **kwargs):
        """Make generator from function and args."""

        n_gen = 0
        while (self.N is None) or (n_gen < self.N):
            obj = f(*args, **kwargs)
            self.history += [obj]
            n_gen += 1
            yield obj

    def __str__(self):
        rval = ', '.join(repr(x) for x in self.history)
        return '%s(N=%s): %s' % (self.__class__.__name__, self.N, rval)


@pytest.fixture(scope='session')
def RandomData():
    """Provide RandomData (helper) class to tests."""
    return _RandomData



"""Utils for managing plugins."""

import importlib
import pkgutil
from pathlib import Path


class PluginError(Exception):
    pass


def detect_model(src):
    """Detects appropriate implementation from a source object
    Return a model object
    """

    if not hasattr(src, 'obj'):
        raise PluginError("Input to detect_model doesn't seem to be a source object")

    plugins = load_plugins()
    detected_classes = [module.Model for module in plugins
                        if hasattr(module, 'Model') and module.Model.detect(src)]

    if len(detected_classes) == 0:
        raise PluginError(f"No support for model {src.obj}")
    if len(detected_classes) > 1:
        raise PluginError(f"More than one model plugin supports model {src.obj}")
    return detected_classes[0]


def load_plugins():
    """Find and import all available plugins
    """
    plugin_path = Path(__file__).resolve().parent

    modules = list()

    for _, modname, ispkg in pkgutil.iter_modules([plugin_path], 'pharmpy.plugins.'):
        if ispkg:
            module = importlib.import_module(modname)
            modules.append(module)

    return modules



from pharmpy.parameter import Parameter, ParameterSet
from pharmpy.parse_utils.generic import AttrToken, remove_token_and_space

from .record import Record

max_upper_bound = 1000000
min_lower_bound = -1000000


class ThetaRecord(Record):
    def parameters(self, first_theta):
        """Get a parameter set for this theta record.
        first_theta is the number of the first theta in this record
        """
        pset = ParameterSet()
        current_theta = first_theta
        for theta in self.root.all('theta'):
            init = theta.init.tokens[0].eval
            fix = bool(theta.find('FIX'))
            if theta.find('low'):
                if theta.low.find('NEG_INF'):
                    lower = min_lower_bound
                else:
                    lower = theta.low.tokens[0].eval
            else:
                lower = min_lower_bound
            if theta.find('up'):
                if theta.up.find('POS_INF'):
                    upper = max_upper_bound
                else:
                    upper = theta.up.tokens[0].eval
            else:
                upper = max_upper_bound
            multiple = theta.find('n')
            if multiple:
                n = multiple.INT
            else:
                n = 1
            for i in range(0, n):
                new_par = Parameter(f'THETA({current_theta})', init, lower, upper, fix)
                current_theta += 1
                pset.add(new_par)
        return pset

    def _multiple(self, theta):
        """Return the multiple (xn) of a theta or 1 if no multiple
        """
        multiple = theta.find('n')
        if multiple:
            n = multiple.INT
        else:
            n = 1
        return n

    def update(self, parameters, first_theta):
        """From a ParameterSet update the THETAs in this record

        Currently only updating initial estimates
        """
        i = first_theta
        for theta in self.root.all('theta'):
            name = f'THETA({i})'
            param = parameters[name]
            new_init = param.init
            if float(str(theta.init)) != new_init:
                theta.init.tokens[0].value = str(new_init)
            fix = bool(theta.find('FIX'))
            if fix != param.fix:
                if param.fix:
                    space = AttrToken('WS', ' ')
                    fix_token = AttrToken('FIX', 'FIX')
                    theta.children.extend([space, fix_token])
                else:
                    remove_token_and_space(theta, 'FIX')

            n = self._multiple(theta)
            i += n

    def __len__(self):
        """Number of thetas in this record
        """
        tot = 0
        for theta in self.root.all('theta'):
            tot += self._multiple(theta)
        return tot

    # @thetas.setter
    # def thetas(self, thetas):
    #    nodes = []
    #    nodes_new = self._new_theta_nodes(thetas)
    #    for child in self.root.children:
    #        if child.rule != 'theta':
    #            nodes += [child]
    #            continue
    #        try:
    #            nodes += [nodes_new.pop(0)]
    #        except IndexError:
    #            pass
    #    self.root = AttrTree.create('root', nodes + nodes_new)

    # def _new_theta_nodes(self, thetas):
    #    nodes = []
    #    for theta in thetas:
    #        if nodes:
    #            nodes += [dict(WS='\n  ')]
    #        new = [{'LPAR': '('}]
    #        if theta.lower is not None:
    #            new += [{'low': {'NUMERIC': theta.lower}}]
    #            new += [{'WS': ' '}]
    #        if theta.init is not None:
    #            new += [{'init': {'NUMERIC': theta.init}}]
    #            new += [{'WS': ' '}]
    #        if theta.upper is not None:
    #            new += [{'up': {'NUMERIC': theta.upper}}]
    #        if theta.fix:
    #            new += [{'WS': ' '}]
    #            new += [{'FIX': 'FIXED'}]
    #        new += [{'RPAR': ')'}]
    #        nodes += [AttrTree.create('theta', new)]
    #    return nodes



"""
Generic NONMEM code record class.

"""

import lark
import sympy

from pharmpy.data_structures import OrderedSet
from pharmpy.statements import Assignment, ModelStatements

from .record import Record


class ExpressionInterpreter(lark.visitors.Interpreter):
    def visit_children(self, tree):
        """Does not visit tokens
        """
        return [self.visit(child) for child in tree.children if isinstance(child, lark.Tree)]

    def expression(self, node):
        t = self.visit_children(node)
        if len(t) > 2:
            op = t[1]
            terms = t[0::2]
            if op == '+':
                expr = sympy.Add(*terms)
            elif op == '-':
                expr = terms[0]
                for term in terms[1:]:
                    expr -= term
            elif op == '*':
                expr = sympy.Mul(*terms)
            elif op == '/':
                expr = terms[0]
                for term in terms[1:]:
                    expr /= term
        else:
            expr = t[0]
        return expr

    def logical_expression(self, node):
        t = self.visit_cildren(node)
        if len(t) > 2:
            left, op, right = self.visit_children(node)
            return op(left, right)
        elif len(t) == 2:
            op, expr = self.visit_children(node)
            return op(expr)
        else:
            return t[0]

    def logical_operator(self, node):
        name = str(node)
        if name == '==' or name == '.EQ.':
            return sympy.Eq
        elif name == '/=' or name == '.NE.':
            return sympy.Ne
        elif name == '<=' or name == '.LE.':
            return sympy.Le
        elif name == '>=' or name == '.GE.':
            return sympy.Ge
        elif name == '<' or name == '.LT.':
            return sympy.Lt
        elif name == '>' or name == '.GT.':
            return sympy.Gt
        elif name == '.AND.':
            return sympy.And
        elif name == '.OR.':
            return sympy.Or
        elif name == '.NOT.':
            return sympy.Not

    def func(self, node):
        func, expr = self.visit_children(node)
        return func(expr)

    def func2(self, node):
        a, p = self.visit_children(node)
        return sympy.Mod(a, p)

    def intrinsic_func(self, node):
        name = str(node)
        if name == "EXP":
            return sympy.exp
        elif name == "LOG":
            return sympy.log
        elif name == "LOG10":
            return lambda x: sympy.log(x, 10)
        elif name == "SQRT":
            return sympy.sqrt
        elif name == "SIN":
            return sympy.sin
        elif name == "COS":
            return sympy.cos
        elif name == "ABS":
            return sympy.Abs
        elif name == "TAN":
            return sympy.tan
        elif name == "ASIN":
            return sympy.asin
        elif name == "ACOS":
            return sympy.acos
        elif name == "ATAN":
            return sympy.atan
        elif name == "INT":
            return lambda x: sympy.sign(x) * sympy.floor(sympy.Abs(x))
        elif name == "GAMLN":
            return sympy.loggamma

    def power(self, node):
        b, e = self.visit_children(node)
        return b**e

    def operator(self, node):
        return str(node)

    def number(self, node):
        try:
            return sympy.Integer(str(node))
        except ValueError:
            return sympy.Float(str(node))

    def symbol(self, node):
        return sympy.Symbol(str(node), real=True)


class CodeRecord(Record):
    @property
    def statements(self):
        s = []
        for statement in self.root.all('statement'):
            node = statement.children[0]
            if node.rule == 'assignment':
                name = str(node.variable)
                expr = ExpressionInterpreter().visit(node.expression)
                ass = Assignment(name, expr)
                s.append(ass)
            elif node.rule == 'logical_if':
                logic_expr = ExpressionInterpreter().visit(node.logical_expression)
                name = str(node.assignment.variable)
                expr = ExpressionInterpreter().visit(node.assignment.expression)
                pw = sympy.Piecewise((expr, logic_expr))
                ass = Assignment(name, pw)
                s.append(ass)
            elif node.rule == 'block_if':
                interpreter = ExpressionInterpreter()
                blocks = []     # [(logic, [(symb1, expr1), ...]), ...]
                symbols = OrderedSet()

                first_logic = interpreter.visit(node.block_if_start.logical_expression)
                first_block = node.block_if_start
                first_symb_exprs = []
                for assign_node in first_block.all('assignment'):
                    name = str(assign_node.variable)
                    first_symb_exprs.append((name, interpreter.visit(assign_node.expression)))
                    symbols.add(name)
                blocks.append((first_logic, first_symb_exprs))

                else_if_blocks = node.all('block_if_elseif')
                for elseif in else_if_blocks:
                    logic = interpreter.visit(elseif.logical_expression)
                    elseif_symb_exprs = []
                    for assign_node in elseif.all('assignment'):
                        name = str(assign_node.variable)
                        elseif_symb_exprs.append((name, interpreter.visit(assign_node.expression)))
                        symbols.add(name)
                    blocks.append((logic, elseif_symb_exprs))

                else_block = node.find('block_if_else')
                if else_block:
                    else_symb_exprs = []
                    for assign_node in else_block.all('assignment'):
                        name = str(assign_node.variable)
                        else_symb_exprs.append((name, interpreter.visit(assign_node.expression)))
                        symbols.add(name)
                    blocks.append((True, else_symb_exprs))

                for symbol in symbols:
                    pairs = []
                    for block in blocks:
                        logic = block[0]
                        for cursymb, expr in block[1]:
                            if cursymb == symbol:
                                pairs.append((expr, logic))
                    pw = sympy.Piecewise(*pairs)
                    ass = Assignment(symbol, pw)
                    s.append(ass)

        return ModelStatements(s)



# The NONMEM Model class
import re
import shutil
from pathlib import Path

import pharmpy.data
import pharmpy.model
from pharmpy.data import DatasetError
from pharmpy.parameter import ParameterSet
from pharmpy.plugins.nonmem.results import NONMEMChainedModelfitResults
from pharmpy.plugins.nonmem.table import NONMEMTableFile, PhiTable
from pharmpy.random_variables import RandomVariables

from .nmtran_parser import NMTranParser


class Model(pharmpy.model.Model):
    def __init__(self, src, **kwargs):
        super().__init__()
        parser = NMTranParser()
        self.source = src
        if not self.source.filename_extension:
            self.source.filename_extension = '.ctl'
        self.name = self.source.path.stem
        self.control_stream = parser.parse(src.code)
        self._parameters_updated = False
        self._initial_individual_estimates_updated = False
        self._updated_etas_file = None
        self._dataset_updated = False

    @property
    def modelfit_results(self):
        try:
            return self._modelfit_results
        except AttributeError:
            None
        if self.source.path.is_file():
            lst_path = self.source.path.with_suffix('.lst')
            if lst_path.exists():
                num_est = len(self.control_stream.get_records('ESTIMATION'))
                self._modelfit_results = NONMEMChainedModelfitResults(lst_path, num_est, model=self)
                return self._modelfit_results
        else:
            return None

    @staticmethod
    def detect(src, *args, **kwargs):
        """ Check if src represents a NONMEM control stream
        i.e. check if it is a file that contain $PRO
        """
        return bool(re.search(r'^\$PRO', src.code, re.MULTILINE))

    def update_source(self, path=None, force=False):
        """ Update the source

            path - path to modelfile
        """
        if self._dataset_updated:
            # FIXME: If no name set use the model name. Set that when setting dataset to input!
            datapath = self.dataset.pharmpy.write_csv(force=force)
            self.path = datapath

            data_record = self.control_stream.get_records('DATA')[0]

            label = self.dataset.columns[0]
            data_record.ignore_character_from_header(label)
            self._update_input(self.input.dataset.columns)

            # Remove IGNORE/ACCEPT. Could do diff between old dataset and find simple
            # IGNOREs to add i.e. for filter out certain ID.
            del(data_record.ignore)
            del(data_record.accept)
            self._dataset_updated = False

        self._update_parameters()
        self._update_initial_individual_etimates(path)

        super().update_source()

    def _update_parameters(self):
        """ Update parameters
        """
        if not self._parameters_updated:
            return

        params = self.parameters
        next_theta = 1
        for theta_record in self.control_stream.get_records('THETA'):
            theta_record.update(params, next_theta)
            next_theta += len(theta_record)
        next_omega = 1
        previous_size = None
        for omega_record in self.control_stream.get_records('OMEGA'):
            next_omega, previous_size = omega_record.update(params, next_omega, previous_size)
        next_sigma = 1
        previous_size = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            next_sigma, previous_size = sigma_record.update(params, next_sigma, previous_size)

        self._parameters_updated = False

    def _update_initial_individual_etimates(self, path):
        """ Update $ETAS

            Could have 0 FIX in model. Need to readd these
        """
        if path is None:        # What to do here?
            phi_path = Path('.')
        else:
            phi_path = path.parent
        phi_path /= f'{self.name}_input.phi'

        if self._initial_individual_estimates_updated:
            etas = self.initial_individual_estimates
            zero_fix = self._zero_fix_rvs(eta=True)
            if zero_fix:
                for eta in zero_fix:
                    etas[eta] = 0
            etas = self._sort_eta_columns(etas)
            phi = PhiTable(df=etas)
            table_file = NONMEMTableFile(tables=[phi])
            table_file.write(phi_path)
            # FIXME: This is a common operation
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS ')
            record.path = phi_path
        elif self._updated_etas_file:
            eta_records = self.control_stream.get_records('ETAS')
            if eta_records:
                record = eta_records[0]
            else:
                record = self.control_stream.append_record('$ETAS')
            shutil.copy(self._updated_etas_file, phi_path)
            record.path = phi_path

        if self._initial_individual_estimates_updated or self._updated_etas_file:
            first_est_record = self.control_stream.get_records('ESTIMATION')[0]
            try:
                first_est_record.option_pairs['MCETA']
            except KeyError:
                first_est_record.set_option('MCETA', 1)
            self._updated_etas_file = None
            self._initial_individual_estimates_updated = False

    def validate(self):
        """Validates NONMEM model (records) syntactically."""
        self.control_stream.validate()

    @property
    def parameters(self):
        """Get the ParameterSet of all parameters
        """
        try:
            return self._parameters
        except AttributeError:
            pass

        next_theta = 1
        params = ParameterSet()
        for theta_record in self.control_stream.get_records('THETA'):
            thetas = theta_record.parameters(next_theta)
            params.update(thetas)
            next_theta += len(thetas)
        next_omega = 1
        previous_size = None
        for omega_record in self.control_stream.get_records('OMEGA'):
            omegas, next_omega, previous_size = omega_record.parameters(next_omega, previous_size)
            params.update(omegas)
        next_sigma = 1
        previous_size = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            sigmas, next_sigma, previous_size = sigma_record.parameters(next_sigma, previous_size)
            params.update(sigmas)
        self._parameters = params
        return params

    @parameters.setter
    def parameters(self, params):
        """params can be a ParameterSet or a dict-like with name: value

           Current restricions:
            * ParameterSet only supported for new initial estimates and fix
            * Only set the exact same parameters. No additions and no removing of parameters
        """
        if isinstance(params, ParameterSet):
            inits = params.inits
        else:
            inits = params
        if not self.random_variables.validate_parameters(inits):
            raise ValueError("New parameter inits are not valid")

        parameters = self.parameters
        if not isinstance(params, ParameterSet):
            parameters.inits = params
        else:
            if len(self.parameters) != len(params):
                raise NotImplementedError("Current number of parameters in model and parameters to "
                                          "set are not the same. Not yet supported")

            for p in self.parameters:
                name = p.name
                if name not in params:
                    raise NotImplementedError(f"Parameter '{name}' not found in input. Currently "
                                              f"not supported")
            self._parameters = params
        self._parameters_updated = True

    @property
    def initial_individual_estimates(self):
        '''Initial individual estimates

           These are taken from the $ETAS FILE. 0 FIX ETAs are removed.
           If no $ETAS is present None will be returned.

           Setter assumes that all IDs are present
        '''
        try:
            return self._initial_individual_estimates
        except AttributeError:
            pass
        etas = self.control_stream.get_records('ETAS')
        if etas:
            path = Path(etas[0].path)
            if not path.is_absolute():
                source_dir = self.source.path.parent
                path = source_dir / path
                path = path.resolve()
            phi_tables = NONMEMTableFile(path)
            rv_names = [rv.name for rv in self.random_variables if rv.name.startswith('ETA')]
            etas = next(phi_tables).etas[rv_names]
            self._initial_individual_estimates = etas
        else:
            self._initial_individual_estimates = None
        return self._initial_individual_estimates

    @initial_individual_estimates.setter
    def initial_individual_estimates(self, estimates):
        rv_names = {rv.name for rv in self.random_variables if rv.name.startswith('ETA')}
        columns = set(estimates.columns)
        if columns < rv_names:
            raise ValueError(f'Cannot set initial estimate for random variable not in the model:'
                             f' {rv_names - columns}')
        diff = columns - rv_names
        # If not setting all etas automatically set remaining to 0 for all individuals
        if len(diff) > 0:
            for name in diff:
                estimates = estimates.copy(deep=True)
                estimates[name] = 0
            estimates = self._sort_eta_columns(estimates)
        self._initial_individual_estimates = estimates
        self._initial_individual_estimates_updated = True
        self._updated_etas_file = None

    def update_individual_estimates(self, source):
        """Update initial individual estimates from another model
        """
        self._initial_individual_estimates_updated = False
        try:
            del self._initial_individual_estimates
        except AttributeError:
            pass
        self._updated_etas_file = self.source.path.with_suffix('.phi')

    def _sort_eta_columns(self, df):
        colnames = df.columns

        def keyfunc(name):
            m = re.match(r'ETA\((\d+)\)', name)
            return int(m.group(1))
        sorted_colnames = colnames.sort(key=keyfunc)
        return df.reindex(sorted_colnames, axis=1)

    @property
    def statements(self):
        pred = self.control_stream.get_records('PRED')
        if len(pred) == 0:
            pk = self.control_stream.get_records('PK')
            rec = pk[0]
        else:
            rec = pred[0]
        return rec.statements

    def _zero_fix_rvs(self, eta=True):
        zero_fix = []
        if eta:
            prev_cov = None
            next_omega = 1
            for omega_record in self.control_stream.get_records('OMEGA'):
                _, next_omega, prev_cov, new_zero_fix = \
                        omega_record.random_variables(next_omega, prev_cov)
                zero_fix += new_zero_fix
        else:
            prev_cov = None
            next_sigma = 1
            for sigma_record in self.control_stream.get_records('SIGMA'):
                _, next_sigma, prev_cov, new_zero_fix = \
                        sigma_record.random_variables(next_sigma, prev_cov)
                zero_fix += new_zero_fix
        return zero_fix

    @property
    def random_variables(self):
        rvs = RandomVariables()
        next_omega = 1
        prev_cov = None
        for omega_record in self.control_stream.get_records('OMEGA'):
            etas, next_omega, prev_cov, _ = omega_record.random_variables(next_omega, prev_cov)
            rvs.update(etas)
        next_sigma = 1
        prev_cov = None
        for sigma_record in self.control_stream.get_records('SIGMA'):
            epsilons, next_sigma, prev_cov, _ = sigma_record.random_variables(next_sigma, prev_cov)
            rvs.update(epsilons)
        return rvs

    def __str__(self):
        return str(self.control_stream)

    @property
    def dataset_path(self):
        """This property is NONMEM specific
        """
        record = self.control_stream.get_records('DATA')[0]
        path = Path(record.filename)
        if not path.is_absolute():
            path = self.source.path.parent / path     # Relative model source file.
        try:
            return path.resolve()
        except FileNotFoundError:
            return path

    @dataset_path.setter
    def dataset_path(self, path):
        path = Path(path)
        assert not path.exists() or path.is_file(), ('input path change, but non-file exists at '
                                                     'target (%s)' % str(path))
        record = self.control_stream.get_records('DATA')[0]
        record.filename = str(path)

    @property
    def dataset(self):
        try:
            return self._data_frame
        except AttributeError:
            self._data_frame = self._read_dataset(raw=False)
        return self._data_frame

    @dataset.setter
    def dataset(self, df):
        self._dataset_updated = True
        self._data_frame = df

    def read_raw_dataset(self, parse_columns=tuple()):
        return self._read_dataset(raw=True, parse_columns=parse_columns)

    @staticmethod
    def _synonym(key, value):
        """Return a tuple reserved name and synonym
        """
        _reserved_column_names = [
            'ID', 'L1', 'L2', 'DV', 'MDV', 'RAW_', 'MRG_', 'RPT_',
            'TIME', 'DATE', 'DAT1', 'DAT2', 'DAT3', 'EVID', 'AMT', 'RATE', 'SS', 'II', 'ADDL',
            'CMT', 'PCMT', 'CALL', 'CONT'
        ]
        if key in _reserved_column_names:
            return (key, value)
        elif value in _reserved_column_names:
            return (value, key)
        else:
            raise DatasetError(f'A column name "{key}" in $INPUT has a synonym to a non-reserved '
                               f'column name "{value}"')

    def _column_info(self):
        """List all column names in order.
            Use the synonym when synonym exists.
            return tuple of three lists, colnames, coltypes and drop together with a dictionary
            of replacements for reserved names (aka synonyms).
            Anonymous columns, i.e. DROP or SKIP alone, will be given unique names _DROP1, ...
        """
        input_records = self.control_stream.get_records("INPUT")
        colnames = []
        coltypes = []
        drop = []
        synonym_replacement = {}
        next_anonymous = 1
        for record in input_records:
            for key, value in record.all_options:
                if value:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        colnames.append(value)
                        reserved_name = value
                    elif value == 'DROP' or value == 'SKIP':
                        colnames.append(key)
                        drop.append(True)
                        reserved_name = key
                    else:
                        drop.append(False)
                        (reserved_name, synonym) = Model._synonym(key, value)
                        synonym_replacement[reserved_name] = synonym
                        colnames.append(synonym)
                else:
                    if key == 'DROP' or key == 'SKIP':
                        drop.append(True)
                        name = f'_DROP{next_anonymous}'
                        colnames.append(name)
                        reserved_name = name
                        next_anonymous += 1
                    else:
                        drop.append(False)
                        colnames.append(key)
                        reserved_name = key
                coltypes.append(pharmpy.data.read.infer_column_type(reserved_name))
        return colnames, coltypes, drop, synonym_replacement

    def _update_input(self, new_names):
        """Update $INPUT with new column names

           currently supporting append columns at end
        """
        colnames, _, _, _ = self._column_info()
        appended_names = new_names[len(colnames):]
        input_records = self.control_stream.get_records("INPUT")
        last_input_record = input_records[-1]
        for colname in appended_names:
            last_input_record.append_option(colname)

    def _replace_synonym_in_filters(filters, replacements):
        result = []
        for f in filters:
            if f.COLUMN in replacements:
                s = ''
                for child in f.children:
                    if child.rule == 'COLUMN':
                        value = replacements[f.COLUMN]
                    else:
                        value = str(child)
                    s += value
            else:
                s = str(f)
            result.append(s)
        return result

    def _read_dataset(self, raw=False, parse_columns=tuple()):
        data_records = self.control_stream.get_records('DATA')
        ignore_character = data_records[0].ignore_character
        null_value = data_records[0].null_value
        (colnames, coltypes, drop, replacements) = self._column_info()

        if raw:
            ignore = None
            accept = None
        else:
            # FIXME: All direct handling of control stream spanning
            # over one or more records should move
            ignore = data_records[0].ignore
            accept = data_records[0].accept
            # FIXME: This should really only be done if setting the dataset
            if ignore:
                ignore = Model._replace_synonym_in_filters(ignore, replacements)
            else:
                accept = Model._replace_synonym_in_filters(accept, replacements)

        df = pharmpy.data.read_nonmem_dataset(self.dataset_path, raw, ignore_character, colnames,
                                              coltypes, drop, null_value=null_value,
                                              parse_columns=parse_columns, ignore=ignore,
                                              accept=accept)
        df.name = self.dataset_path.stem
        return df



import sympy
from pyfakefs.fake_filesystem_unittest import Patcher
from sympy import Symbol

from pharmpy import Model
from pharmpy.parameter import Parameter


def test_source(pheno_path):
    model = Model(pheno_path)
    assert model.source.code.startswith(';; 1.')


def test_update_inits(pheno_path):
    model = Model(pheno_path)
    model.update_inits()


def test_parameters(pheno_path):
    model = Model(pheno_path)
    params = model.parameters
    assert len(params) == 6
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 0.00469307, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 1.00916, lower=0, upper=1000000)
    assert model.parameters['THETA(3)'] == Parameter('THETA(3)', 0.1, lower=-0.99, upper=1000000)
    assert model.parameters['OMEGA(1,1)'] == Parameter('OMEGA(1,1)', 0.0309626,
                                                       lower=0, upper=sympy.oo)
    assert model.parameters['OMEGA(2,2)'] == Parameter('OMEGA(2,2)', 0.031128,
                                                       lower=0, upper=sympy.oo)
    assert model.parameters['SIGMA(1,1)'] == Parameter('SIGMA(1,1)', 0.0130865,
                                                       lower=0, upper=sympy.oo)


def test_set_parameters(pheno_path):
    model = Model(pheno_path)
    params = {'THETA(1)': 0.75, 'THETA(2)': 0.5, 'THETA(3)': 0.25,
              'OMEGA(1,1)': 0.1, 'OMEGA(2,2)': 0.2, 'SIGMA(1,1)': 0.3}
    model.parameters = params
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 0.75, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 0.5, lower=0, upper=1000000)
    assert model.parameters['THETA(3)'] == Parameter('THETA(3)', 0.25, lower=-0.99, upper=1000000)
    assert model.parameters['OMEGA(1,1)'] == Parameter('OMEGA(1,1)', 0.1,
                                                       lower=0, upper=sympy.oo)
    assert model.parameters['OMEGA(2,2)'] == Parameter('OMEGA(2,2)', 0.2,
                                                       lower=0, upper=sympy.oo)
    assert model.parameters['SIGMA(1,1)'] == Parameter('SIGMA(1,1)', 0.3,
                                                       lower=0, upper=sympy.oo)
    model.update_source()
    thetas = model.control_stream.get_records('THETA')
    assert str(thetas[0]) == '$THETA  (0,0.75) ; CL\n'
    assert str(thetas[1]) == '$THETA  (0,0.5) ; V\n'
    assert str(thetas[2]) == '$THETA  (-.99,0.25)\n'
    omegas = model.control_stream.get_records('OMEGA')
    assert str(omegas[0]) == '$OMEGA  DIAGONAL(2)\n 0.1  ;       IVCL\n 0.2  ;        IVV\n'
    sigmas = model.control_stream.get_records('SIGMA')
    assert str(sigmas[0]) == '$SIGMA  0.3\n'

    model = Model(pheno_path)
    params = model.parameters
    params['THETA(1)'].init = 18
    model.parameters = params
    assert model.parameters['THETA(1)'] == Parameter('THETA(1)', 18, lower=0, upper=1000000)
    assert model.parameters['THETA(2)'] == Parameter('THETA(2)', 1.00916, lower=0, upper=1000000)


def test_results(pheno_path):
    model = Model(pheno_path)
    assert len(model.modelfit_results) == 1     # A chain of one estimation


def test_minimal(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    assert len(model.statements) == 1
    assert model.statements[0].expression == \
        Symbol('THETA(1)', real=True) + Symbol('ETA(1)', real=True) + Symbol('ERR(1)', real=True)


def test_copy(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    copy = model.copy()
    assert id(model) != id(copy)
    assert model.statements[0].expression == \
        Symbol('THETA(1)', real=True) + Symbol('ETA(1)', real=True) + Symbol('ERR(1)', real=True)


def test_initial_individual_estimates(datadir):
    path = datadir / 'minimal.mod'
    model = Model(path)
    assert model.initial_individual_estimates is None

    path = datadir / 'pheno_etas.mod'
    model = Model(path)
    inits = model.initial_individual_estimates
    assert len(inits) == 59
    assert len(inits.columns) == 2
    assert inits['ETA(1)'][2] == -0.166321


def test_update_individual_estimates(datadir):
    with Patcher(additional_skip_names=['pkgutil']) as patcher:
        fs = patcher.fs
        fs.add_real_file(datadir / 'pheno_real.mod', target_path='run1.mod')
        fs.add_real_file(datadir / 'pheno_real.phi', target_path='run1.phi')
        fs.add_real_file(datadir / 'pheno_real.lst', target_path='run1.lst')
        fs.add_real_file(datadir / 'pheno_real.ext', target_path='run1.ext')
        model = Model('run1.mod')
        model.name = 'run2'
        model.update_individual_estimates(model)
        model.update_source()
        with open('run2_input.phi', 'r') as fp, open('run1.phi') as op:
            assert fp.read() == op.read()
        assert str(model).endswith("""$ESTIMATION METHOD=1 INTERACTION PRINT=1 MCETA=1
$COVARIANCE UNCONDITIONAL
$TABLE      ID TIME AMT WGT APGR IPRED PRED TAD CWRES NPDE NOAPPEND
            NOPRINT ONEHEADER FILE=pheno_real.tab
$ETAS FILE=run2_input.phi""")



import sympy

from pharmpy import Model
from pharmpy.statements import Compartment, Elimination, IVAbsorption


def S(x):
    return sympy.Symbol(x, real=True)


def test_subs(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    statements = model.statements

    statements.subs(S('ETA(1)'), S('ETAT1'))

    assert statements[5].expression == S('TVCL') * sympy.exp(S('ETAT1'))


def test_Compartment():
    absorption = IVAbsorption('AMT')
    CL = S('CL')
    V = S('V')
    elimination = Elimination(CL / V)
    comp = Compartment('central', absorption, elimination)
    assert comp



import pytest
import sympy

from pharmpy.parameter import Parameter, ParameterSet


@pytest.mark.parametrize('name,init,lower,upper,fix', [
    ('THETA(1)', 23, None, None, None),
    ('X', 12, None, None, None),
    ('_NAME', 0, None, None, None),
    ('OMEGA(2,1)', 0.1, 0, None, None),
    ('TCVL', 0.23, -2, 2, None),
    ])
def test_initialization(name, init, lower, upper, fix):
    param = Parameter(name, init, lower, upper, fix)
    assert param.name == name
    assert param.init == init
    if lower is not None:
        assert param.lower == lower
    else:
        assert param.lower == -sympy.oo
    if upper is not None:
        assert param.upper == upper
    else:
        assert param.upper == sympy.oo
    assert param.fix == bool(fix)


@pytest.mark.parametrize('name,init,lower,upper,fix', [
    ('OMEGA(2,1)', 0.1, 2, None, None),
    ('X', 1, 0, -1, None),
    ])
def test_illegal_initialization(name, init, lower, upper, fix):
    with pytest.raises(ValueError):
        Parameter(name, init, lower, upper, fix)


def test_unconstrain():
    param = Parameter('X', 2, lower=0, upper=23)
    param.unconstrain()
    assert param.lower == -sympy.oo
    assert param.upper == sympy.oo

    fixed_param = Parameter('Y', 0, fix=True)
    fixed_param.unconstrain()
    assert fixed_param.lower == -sympy.oo
    assert fixed_param.upper == sympy.oo


def test_fix():
    param = Parameter('X', 2, lower=0, upper=23)
    assert param == Parameter('X', 2, lower=0, upper=23, fix=False)
    param.fix = True
    assert param == Parameter('X', 2, lower=0, upper=23, fix=True)


def test_init():
    param = Parameter('X', 2, lower=0, upper=23)
    with pytest.raises(ValueError):
        param.init = -1
    param.init = 22


def test_pset_index():
    p = Parameter('Y', 9)
    pset = ParameterSet((p,))
    assert len(pset) == 1
    assert pset['Y'] is p

    p2 = Parameter('Z', 5)
    pset.add(p2)

    assert len(pset) == 2

    # Check that the parameter set keeps the insertion order upon iteration
    for i, param in enumerate(pset):
        if i == 0:
            assert param is p
        else:
            assert param is p2


def test_pset_remove_fixed():
    p1 = Parameter('Y', 9, fix=False)
    p2 = Parameter('X', 3, fix=True)
    p3 = Parameter('Z', 1, fix=False)
    pset = ParameterSet([p1, p2, p3])
    pset.remove_fixed()
    assert len(pset) == 2
    assert pset['Y'] == Parameter('Y', 9)


def test_pset_inits():
    p1 = Parameter('Y', 9)
    p2 = Parameter('X', 3)
    p3 = Parameter('Z', 1)
    pset = ParameterSet([p1, p2, p3])
    pset.inits = {'X': 28}
    assert len(pset) == 3
    assert pset['X'] == Parameter('X', 28)
    assert pset['Y'] == Parameter('Y', 9)
    assert pset['Z'] == Parameter('Z', 1)

    with pytest.raises(KeyError):
        pset.inits = {'CL': 0}

    pset.inits = {'X': 0, 'Y': 2, 'Z': 5}
    assert len(pset) == 3
    assert pset['X'] == Parameter('X', 0)
    assert pset['Y'] == Parameter('Y', 2)
    assert pset['Z'] == Parameter('Z', 5)


def test_pset_fix():
    p1 = Parameter('Y', 9, fix=False)
    p2 = Parameter('X', 3, fix=True)
    p3 = Parameter('Z', 1, fix=False)
    pset = ParameterSet([p1, p2, p3])
    assert pset.fix == {'Y': False, 'X': True, 'Z': False}
    fixedness = {'Y': True, 'X': True, 'Z': True}
    pset.fix = fixedness
    assert pset.fix == {'Y': True, 'X': True, 'Z': True}


def test_pset_repr():
    p1 = Parameter('Y', 9, fix=False)
    pset = ParameterSet([p1])
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str
    pset = ParameterSet()
    assert type(repr(pset)) == str
    assert type(pset._repr_html_()) == str


def test_parameter_space():
    p1 = Parameter('Y', 9, fix=True)
    assert p1.parameter_space == sympy.FiniteSet(9)
    p2 = Parameter('X', 10, lower=0, upper=15)
    assert p2.parameter_space == sympy.Interval(0, 15)



import logging
import os
import re
from glob import glob
from pathlib import Path

from pharmpy.execute import Engine

# Warning: This is experimental and perhaps the architecture needs to change


class NONMEM7(Engine):
    """NONMEM7 execution engine.

    An implementation of the generic :class:`~pharmpy.execute.engine.Engine` (see for more
    information).

    Will automatically scan for installed NONMEM versions (see :func:`~self.scan_installed`).
    """

    def __init__(self, model, version=None):
        log = logging.getLogger(__name__)

        self.installed = self.scan_installed()
        if not version and self.installed:
            version = max(self.installed.keys())
        if version not in self.installed:
            log.error('Requested NONMEM7 version %r not available (found: %s)', version,
                      ', '.join(self.installed.values()))
            self.info = None
        else:
            self.info = (version, *self.installed[version])

        super().__init__(model)

    def get_commandline(self, task):
        if task in {'evaluate', 'estimate'}:
            mod = str(self.model.path)
            lst = re.sub(r'\.(mod|ctl)$', '.lst', mod)
            if mod == lst:
                lst += '.lst'
            return (self.bin, mod, lst)
        else:
            raise ValueError('Engine (%s) does not support task %r.' % (self.__class__.__name__,
                                                                        task))

    @property
    def bin(self):
        return self.info[2] if self.info else None

    @property
    def version(self):
        return self.info[0] if self.info else None

    def __bool__(self):
        return bool(self.installed)

    @classmethod
    def scan_installed(cls):
        log = logging.getLogger(__name__)
        log.info("Initiating NM version scan in filesystem")

        if os.name == 'nt':
            globpaths = ['C:/nm7*', 'C:/nm_7*', 'C:/NONMEM/nm7*', 'C:/NONMEM/nm_7*']
            globpaths = [str(Path(globpath).resolve().expanduser()) for globpath in globpaths]
            script = 'nmfe7%s.bat'
        else:
            globpaths = []
            for root in [Path(x) for x in ('/opt', '~')]:
                tails = ['nm7*', 'nm_7*', 'nm7*', 'nm_7*', 'nonmem/nm7*', 'nonmem/nm_7*']
                globpaths += [root / x for x in tails]
            globpaths = [str(Path(globpath).expanduser()) for globpath in globpaths]
            script = 'nmfe7%s'
        candidates = [path for globpath in globpaths for path in glob(globpath)]

        nm_info = dict()
        for nm_dir in candidates:
            path = Path(nm_dir)

            if not path.is_dir():
                continue
            subdirs = ['run', 'util']
            if not any((path/subdir).is_dir() for subdir in subdirs):
                continue

            ver_suffix = {7.1: '', 7.2: '2', 7.3: '3', 7.4: '4', 7.5: '5'}
            for version, suffix in ver_suffix.items():
                nm_bin = [path/subdir/(script % suffix) for subdir in subdirs]
                nm_bin = list(filter(lambda x: x.is_file(), nm_bin))
                if nm_bin:
                    nm_info[version] = (nm_dir, str(nm_bin[0]))
                    break

        return nm_info



import re

from pharmpy.model import ModelSyntaxError

from .records.factory import create_record
from .records.raw_record import RawRecord


class NMTranParser:
    """Parser for NMTran control streams
    """
    def parse(self, text):
        stream = NMTranControlStream()

        record_strings = re.split(r'^([ \t]*\$)', text, flags=re.MULTILINE)
        first = record_strings.pop(0)       # Empty if nothing before first record
        if first:
            stream.records.append(RawRecord(first))

        for separator, s in zip(record_strings[0::2], record_strings[1::2]):
            record = create_record(separator + s)
            stream.records.append(record)

        return stream


class NMTranControlStream:
    """Representation of a parsed control stream (model file)
    """
    def __init__(self):
        self.records = []
        self._active_problem = 0

    def get_records(self, name):
        """Return a list of all records of a certain type in the current $PROBLEM
        """
        current_problem = -1
        found = []
        for record in self.records:
            if record.name == 'PROBLEM':
                current_problem += 1
            if current_problem == self._active_problem and record.name == name:
                found.append(record)
        return found

    def append_record(self, content):
        """ Create and append record at the end
        """
        record = create_record(content)
        self.records.append(record)
        return record

    def validate(self):
        in_problem = False
        for record in self.records:
            if in_problem and record.name == 'SIZES':
                raise ModelSyntaxError('The SIZES record must come before the first PROBLEM record')
            elif record.name == 'PROBLEM':
                in_problem = True
            if hasattr(record, 'validate'):
                record.validate()

    def __str__(self):
        return ''.join(str(x) for x in self.records)
